/*******************************************************

	FILE: CLClient.c

	DESC: This file contains the main function and the
		primary processing functions for CLClient.
 

	Copyright 1998-2002 SourceGear Corporation.
	All Rights Reserved.


*******************************************************/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#ifdef WIN32
#include <direct.h>
#endif /* WIN32 */

#ifdef UNIX
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#endif /* UNIX */

#ifndef BUILD_STANDARD
#include "CLcrypto.h"
#endif /* !BUILD_STANDARD */

#include "CLcommon.h"
#include "CLClient.h"
#include "CLnet.h"
#include "CLparams.h"
#include "CLTree.h"

/* Function prototypes local to this file */
int Login(int nSocket, RequestInfo *pReqInfo);
int ExecuteCommand(int nSocket, CLTreeNode** pRoot, RequestInfo *pReqInfo, short bLoadTreeFromServer);
int GetDatabases(int nInSocket, RequestInfo *pReqInfo);
int CheckFileStatus(int nSocket, RequestInfo* pInReqInfo, CLTreeNode** ppRoot, char* pszProject, char* pszFile, char* pszWorkDir);
int GetProjectFiles(int nInSocket, CLTreeNode** ppRoot, RequestInfo* pInReqInfo, enum MsgNum eAction);
int ShareFile(int nInSocket, CLTreeNode** ppRoot, RequestInfo *pReqInfo);
int BranchFile(int nInSocket, CLTreeNode** ppRoot, RequestInfo *pReqInfo);
int GetInfo(int nInSocket, CLTreeNode** ppRoot, RequestInfo *pInReqInfo, enum MsgNum nInRespMsgNum, short bPrintInfo);

#define MSG_NO_FILE   0
#define MSG_WITH_FILE 1

/*----------------------------------------------------------
	main() inputs the gaggle of parameters necessary to make
	a connection and specify a full source safe command.  It does
	the following:
		1. parses the parameters
		2. verifies their correctness
		3. opens the connection to the server
		4. logins in to the server
		5. executes the input command
		6. closes the connection to the server
------------------------------------------------------------*/
int main(int argc, char **argv)
{
	int nSocket;			/* The socket for server communication */
	RequestInfo theRequest;	/* The input request */
	int nRetVal;			/* generic return value */
	CLTreeNode* pRoot = NULL;	/* root of the SOS Tree */
	CLTreeNode* pNode = NULL;
	short bLoadTreeFromServer = FALSE;
	FILE* fp = NULL;


	/* Parse the parameters and put into a struct */
	ParseParameters(argc, argv, &theRequest);

	if (!theRequest.szSOSHome)
		theRequest.bAlwaysDownload = 1;
	
	if (theRequest.bVerbose)
	{
		printf("SourceOffSite Command Line Client: %s\n", CLC_VERSION);
	}

	if (VerifyRequestParams(&theRequest) != 0)
	{
		return -1;
	}

	/* load up a sos file */
	if (theRequest.szSOSHome && (theRequest.eCommand != E_GetDatabases))
	{
		pRoot = LoadTreeFromFile(theRequest.szSOSHome, theRequest.szServer, theRequest.szDatabase, &fp, &bLoadTreeFromServer);

		if (!pRoot && bLoadTreeFromServer)
			pRoot = TreeNodeWithName("$/");
	}

	if ((theRequest.szWorkdir == NULL) && (pRoot != NULL))
	{
		if (theRequest.eCommand == E_ShareFile)
		{
			pNode = GetTreeNodeFromProject(pRoot, theRequest.szShareTo);
		} else {
			if (theRequest.szProject == NULL)		/* Default to root project */
				pNode = GetTreeNodeFromProject(pRoot, (const char *)&ROOT_PROJECT);
			else
				pNode = GetTreeNodeFromProject(pRoot, theRequest.szProject);
		}
		theRequest.szWorkdir = GetWorkingDirectoryFromNode(pNode);
	}

#if DEBUG
	{
		printf("Input Parameters:\n\tCommand = %d %s\n", theRequest.eCommand, theRequest.szCommand);
		printf("\tServer = %s\n",	theRequest.szServer);
		printf("\tPort = %d\n",		theRequest.nPort);
		printf("\tDatabase = %s\n", theRequest.szDatabase);
		printf("\tName = %s\n",		theRequest.szName);
		printf("\tPassword = %s\n", theRequest.szPassword);
		printf("\tProject = %s\n",	theRequest.szProject);
		printf("\tFile = %s\n",		theRequest.szFile);
		printf("\tRevision = %s\n", theRequest.szRevision);
		printf("\tLog = %s\n",		theRequest.szLog);
		printf("\tWorkdir = %s\n",	theRequest.szWorkdir);
		printf("\tLabel = %s\n",	theRequest.szLabel);
		printf("\tVerbose = %d\n",	theRequest.bVerbose);
		printf("\tSOSHome = %s\n",	theRequest.szSOSHome);
		printf("\tUseCompression = %d\n", theRequest.bUseCompression);
		printf("\tWriteable = %d\n", theRequest.bWriteable);
		printf("\tShareTo = %s\n", theRequest.szShareTo);
	}
#endif /* DEBUG */

	/* Connect to the server */
	nSocket = Net_ConnectToServer(theRequest.szServer, theRequest.nPort);
	if (nSocket < 0)
	{
		printf("Could not connect to server %s at port %d. Exiting...\n", theRequest.szServer, theRequest.nPort);
		return -1;
	}

	if (theRequest.bVerbose)
	{
		printf("Connected to server %s at port %d.\n", theRequest.szServer, theRequest.nPort);
	}


	/* Process the input command */
	if (theRequest.eCommand != E_GetDatabases)
	{
		/* Login to the server */
		nRetVal = Login(nSocket, &theRequest);
		if (nRetVal >= 0)
		{

			/* Login completed successfully */
			if (theRequest.bVerbose)
			{
				printf("Successfully logged in.\n");
				printf("Sending %s Command to server.\n", theRequest.szCommand);
			}

			/* Execute the input command */
			nRetVal = ExecuteCommand(nSocket, &pRoot, &theRequest, bLoadTreeFromServer);
		}
	} else {
		/*
			Get databases is handled differently from all other
			commands because it does not require a login - the 
			databases are returned as the message body in the 
			initial challenge response from the server
		*/
		nRetVal = GetDatabases(nSocket, &theRequest);

	}

	if (fp)
	{
		SaveTreeToFile(pRoot, NULL, fp);
	}

	/* Close Connection to Server */
	Net_CloseServerConnection(nSocket, &theRequest);
	if (theRequest.bVerbose)
	{
		printf("Closed connection to server %s at port %d.\n", theRequest.szServer, theRequest.nPort);
	}



	return nRetVal;
}

/*----------------------------------------------
	Login to server connected at the input socket.
	Sends the appropriate request messages to the server
	and checks for corresponding responses from server
	to succesfully log the user in to the source safe
	database.  Uses the RequestInfo struct to build
	the appropriate messages.

	Returns: 0 on success, negative number on error
----------------------------------------------*/
int Login(
	int nSocket,			/* A socket connected to a SS server */
	RequestInfo *pReqInfo	/* Request info for building resp. messages */
	)
{
	MessageInfo	serverMsg;		/* server response message */
	int nRetVal;				/* generic return value */
	char *szProtVer;

#ifndef BUILD_STANDARD
	unsigned char *pSessionKey;
	char szKeyLen[4];
	char szChallengeVal[17];
#endif /* !BUILD_STANDARD */

	MI_Initialize(&serverMsg);

	/* Get the expected challenge as the first response
		message by the server
	*/
	nRetVal = Net_GetEntireResponse(nSocket,&serverMsg, pReqInfo);
	if (nRetVal < 0)
	{
		return nRetVal;
	}

	szProtVer = GetHeaderValue(&serverMsg.daHeaders, H_PROTOCOL);
	if (atof(szProtVer) < atof(PROT_VERSION))
	{
		printf("Incompatible protocol version: Expected %s, Received %s", PROT_VERSION, szProtVer);
		return -1;
	}


	switch (serverMsg.eMsgNum)
	{
		case E_Challenge :
			if (pReqInfo->bVerbose)
			{
				printf("Received Challenge from server\n");
			}
			
			/* Create login message */
			MI_Reset(&serverMsg);
			serverMsg.eMsgNum = E_Login;
			DA_Assign(&serverMsg.daMsgText, MSG_LOGIN, -1);
			AddHeader(&serverMsg.daHeaders, H_PROTOCOL, PROT_VERSION);
			AddHeader(&serverMsg.daHeaders, H_DATABASE, pReqInfo->szDatabase);
			AddHeader(&serverMsg.daHeaders, H_USERNAME, pReqInfo->szName);
			AddHeader(&serverMsg.daHeaders, H_PASSWORD, pReqInfo->szPassword);
			
			/* Send the login request message */
			nRetVal = Net_SendRequest(nSocket, &serverMsg, pReqInfo);
			if (nRetVal < 0)
			{
				return nRetVal;
			}
			
			/* Get the response from the server */
			nRetVal = Net_GetEntireResponse(nSocket,&serverMsg, pReqInfo);
			if (nRetVal < 0)
			{
				return nRetVal;
			}
			
			/* If the response is not a Success, then the login failed, so print
			out the response
			*/
			if (serverMsg.eMsgNum == E_AuthFailed)
			{
				printf("Login Failed.  Could not authenticate user name and password.\n");
				return -1;
			} 
			else  if (serverMsg.eMsgNum != E_Success)
			{
				printf("Unexpected server message after login.  Expected a 300 Success, but got a %d (%s)\n", 
					serverMsg.eMsgNum, ConvertMsgNum2MsgStr(serverMsg.eMsgNum));
				printf("Server Message Body = %s.\n", serverMsg.szMsgBody);
				return -1;
			}
			else
			{
				
				return 0;
			}
			break;
#ifndef BUILD_STANDARD
		case E_SecureChallenge :

			if (pReqInfo->szSOSHome == NULL)
			{
				printf("Unable to find key file: no SOSHome directory specified\n");
				return -1;
			}

			pReqInfo->tsTimeStamps.challengeClock = clock();
			pReqInfo->tsTimeStamps.challengeTime = time(NULL);

			if (pReqInfo->bVerbose)
			{
				printf("Received Secure Challenge from server\n");
			}

			if ((GetUserKey(pReqInfo)) != 0)
			{
/*				printf("Unable to open user key file\n"); */
				return -1;
			}

			sprintf(szKeyLen, "%d", pReqInfo->nKeyLength);

			memcpy(szChallengeVal, serverMsg.szMsgBody, pReqInfo->nKeyLength);
			szChallengeVal[pReqInfo->nKeyLength] = 0;

			pReqInfo->bUseEncrypt = 0;

			MI_Reset(&serverMsg);
			serverMsg.eMsgNum = E_Login;
			DA_Assign(&serverMsg.daMsgText, MSG_LOGIN, -1);
			AddHeader(&serverMsg.daHeaders, H_PROTOCOL, PROT_VERSION);
			AddHeader(&serverMsg.daHeaders, H_USERNAME, pReqInfo->szName);
			AddHeader(&serverMsg.daHeaders, H_KEY_LEN, szKeyLen);
			
			/* Send the login request message */
			nRetVal = Net_SendRequest(nSocket, &serverMsg, pReqInfo);
			if (nRetVal < 0)
			{
				return nRetVal;
			}

			
			pReqInfo->bUseEncrypt = 1;

			InitCrypto(&pReqInfo->tsTimeStamps);

			pSessionKey = (unsigned char *)malloc(pReqInfo->nKeyLength);

			GetRandomNumber(pSessionKey, pReqInfo->nKeyLength);

			MI_Reset(&serverMsg);
			serverMsg.eMsgNum = E_Login;
			DA_Assign(&serverMsg.daMsgText, MSG_LOGIN, -1);
			AddHeader(&serverMsg.daHeaders, H_PROTOCOL, PROT_VERSION);
			AddHeader(&serverMsg.daHeaders, H_USERNAME, pReqInfo->szName);
			AddHeader(&serverMsg.daHeaders, H_PASSWORD, pReqInfo->szPassword);
			AddHeader(&serverMsg.daHeaders, H_DATABASE, pReqInfo->szDatabase);
			AddHeader(&serverMsg.daHeaders, H_KEY_LEN, szKeyLen);
			
			serverMsg.szMsgBody = (char *)malloc(pReqInfo->nKeyLength * 2);
			memcpy(serverMsg.szMsgBody, szChallengeVal, pReqInfo->nKeyLength);
			memcpy(serverMsg.szMsgBody + pReqInfo->nKeyLength, pSessionKey, pReqInfo->nKeyLength);
/*			serverMsg.szMsgBody[pReqInfo->nKeyLength * 2] = 0; */
			serverMsg.nBodySize = pReqInfo->nKeyLength * 2;

			/* Send the login request message */
			nRetVal = Net_SendRequest(nSocket, &serverMsg, pReqInfo);
			if (nRetVal < 0)
			{
				return nRetVal;
			}

			pReqInfo->bUseEncrypt = 0;

			free(pReqInfo->bfKey);
			pReqInfo->bfKey = SetKey(pSessionKey, pReqInfo->nKeyLength);

			/* Get the response from the server */
			nRetVal = Net_GetEntireResponse(nSocket,&serverMsg, pReqInfo);
			if (nRetVal < 0)
			{
				return nRetVal;
			}
			
			pReqInfo->bUseEncrypt = 1;

			/* If the response is not a Success, then the login failed, so print
			out the response
			*/
			if (serverMsg.eMsgNum == E_AuthFailed)
			{
				printf("Login Failed.  Could not authenticate user name and password.\n");
				return -1;
			} 
			else  if (serverMsg.eMsgNum != E_Success)
			{
				printf("Unexpected server message after login.  Expected a 300 Success, but got a %d (%s)\n", 
					serverMsg.eMsgNum, ConvertMsgNum2MsgStr(serverMsg.eMsgNum));
				printf("Server Message Body = %s.\n", serverMsg.szMsgBody);
				return -1;
			}
			else
			{
				
				return 0;
			}

			break;
#endif /* !BUILD_STANDARD */
		default :
			printf("Unexpected server message.  Expected a 200 Challenge, but got a %d (%s)\n", 
				serverMsg.eMsgNum, ConvertMsgNum2MsgStr(serverMsg.eMsgNum));
			return -1;
	}


}


/*----------------------------------------------
	Create a server message (minus the body) from the
	request info structure input.

	Returns: Nothing
----------------------------------------------*/
void CreateServerMessage(
	enum MsgNum	nInMsgNum,		/* Msg num to use (can be diff from  */
								/* one in the req info struct)	*/
	RequestInfo *pInReqInfo,	/* Request info to pull headers from */
	MessageInfo	*pOutServerMsg /* server message to create */
	)
{
	DArray *daHeaders = &pOutServerMsg->daHeaders;

	/* Initialize the Message Info variables */
	MI_Initialize(pOutServerMsg);

	/* Add the message number to the server message */
	pOutServerMsg->eMsgNum = nInMsgNum;


	/* For each parameter in the request that matches
	   a header for a non login command, add the header 
	   to header array.
	*/

	/* Current-Project Header */
	if (pInReqInfo->szProject != NULL)
	{
		AddHeader(daHeaders, H_CUR_PROJECT, pInReqInfo->szProject);
	}

	/* File Name Header */
	if (pInReqInfo->szFile != NULL)
	{
		AddHeader(daHeaders, H_FILE_NAME, ExtractFileName(pInReqInfo->szFile));
	}
	
	/* Revision Header */
	if (pInReqInfo->szRevision != NULL)
	{
		AddHeader(daHeaders, H_REVISION, pInReqInfo->szRevision);
	}

	/* Log Header */
	if (pInReqInfo->szLog != NULL)
	{
		AddHeader(daHeaders, H_LOG, pInReqInfo->szLog);
	}

	/* Label Header */
	if (pInReqInfo->szLabel != NULL)
	{
		AddHeader(daHeaders, H_LABEL, pInReqInfo->szLabel);
	}

	/* Recursive Header */
	if (pInReqInfo->bRecursive != 0)
	{
		AddHeader(daHeaders, H_RECURSIVE, "Yes");
	}

	/* Permanent Header */
	if (pInReqInfo->bPermanent != 0)
	{
		AddHeader(daHeaders, H_PERMANENT, "Yes");
	}

	/* New Name Header (gets mapped to NAME) */
	if (pInReqInfo->szNewName != NULL)
	{
		AddHeader(daHeaders, H_NAME, pInReqInfo->szNewName);
	}

	/* Use Compression Header */
	if (pInReqInfo->bUseCompression != 0)
	{
		AddHeader(daHeaders, H_USE_COMPRESSION, "YES");
		pOutServerMsg->bUseCompression = 1;
	}

	/* Share Project Header */
	if (pInReqInfo->szShareTo != NULL)
	{
		AddHeader(daHeaders, H_SHAREPROJECT, pInReqInfo->szShareTo);
	}

	/* Branch After Share Header */
	if (pInReqInfo->bBranchAfterShare != 0)
	{
		AddHeader(daHeaders, H_BRANCH_AFTER_SHARE, "YES");
	}

}

/*----------------------------------------------
	Send a single request to the server and return
	the single response to that request.  Should
	not be used for requests that can return
	multiple responses.

	Returns: Nothing
----------------------------------------------*/
int SendRequestAndGetResponse(
	int			nInSocket,		/* an open socket connected to a SS Server */
	enum MsgNum	nInMsgNum,		/* The message num for the request */
	RequestInfo *pInReqInfo,	/* Request to send to server */
	int			bInSendFile,	/* Whether to attach the file specified in the */
								/* request to the server.  Use either MSG_NO_FILE or MSG_WITH_FILE */
	int			bInReceiveFile,	/* Whether a file is expected to be returned by the server */
								/* Use either MSG_NO_FILE or MSG_WITH_FILE */
	MessageInfo *pOutResponse	/* Response returned by server */

	)
{
	int nRetVal;				/* Generic return value */
	MessageInfo reqMsg;			/* The request message to send to server */
	int nMsgID=0;				/* Message ID of the request */
	int nRespToID;				/* The value of the response-to header */


	/* Ensure the response message is initialized */
	MI_Initialize(pOutResponse);

	/* Create the server request message */
	CreateServerMessage(nInMsgNum, pInReqInfo, &reqMsg);

	/* Send the request */
	if (bInSendFile)
	{
		/* If the request sends a file, append it to the request */
		nRetVal = Net_SendRequestAppendFile(nInSocket, &reqMsg, 
			pInReqInfo->szFile, pInReqInfo->szWorkdir, pInReqInfo);
	} else {
		/* Send the request as is */
		nRetVal = Net_SendRequest(nInSocket, &reqMsg, pInReqInfo);
	}

	if (nRetVal < 0)
	{
		return nRetVal;
	}

	/* Save the message ID to verify the response is to this message */
	nMsgID = GetIntHeaderValue(&reqMsg.daHeaders, H_MESSAGE_ID);
	if (nMsgID < 0)
	{
		printf("Internal Error:  No MessageID associated with server request!\n");
		return -1;
	}

	/* Get the response from the server.  If the response sends a file
		back, dump it to the location indicated.
	*/
	if (bInReceiveFile)
	{
		nRetVal = Net_GetResponse(nInSocket, pOutResponse, pInReqInfo,
			pInReqInfo->szFile, pInReqInfo->szWorkdir, 1);

	} else {
		/* Get the entire response  */
		nRetVal = Net_GetEntireResponse(nInSocket,pOutResponse, pInReqInfo);
	}

	/* If the response-to header of the response does not match
		the message ID sent to the server, then this message is
		in response to a different message sent to the server, so
		keep reading until we get the message we are looking for
	*/
	nRespToID = GetIntHeaderValue(&pOutResponse->daHeaders, H_RESPONSE_TO);
	while (nRetVal >= 0 && nMsgID != nRespToID)
	{
		/* If there was no ResponseTo header value, then we are out of
			sync with the server, so break */
		if (nRespToID < 0)
		{
			break;
		}

		if (bInReceiveFile)
		{
			nRetVal = Net_GetResponse(nInSocket, pOutResponse, pInReqInfo,
				pInReqInfo->szFile, pInReqInfo->szWorkdir, 1);

		} else {
			/* Get the entire response  */
			nRetVal = Net_GetEntireResponse(nInSocket,pOutResponse, pInReqInfo);
		}
		nRespToID = GetIntHeaderValue(&pOutResponse->daHeaders, H_RESPONSE_TO);

	}

	return nRetVal;

}

/*----------------------------------------------
	Sends message to server to create a SourceSafe
	project, and handles the response.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int CreateProject(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,    /* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	MessageInfo	respMsg;		/* Response Message to get from server */
	int nRetVal;				/* generic return value */

	/* Send the request message to the server.*/
	nRetVal = SendRequestAndGetResponse(nInSocket, E_CreateProject,
				pInReqInfo, MSG_NO_FILE, MSG_NO_FILE, &respMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}

	/* Print results */
	switch (respMsg.eMsgNum)
	{
		case E_Success :
			{
				if (*ppRoot)
				{
					GetTreeNodeFromProject(*ppRoot, pInReqInfo->szProject);
				}

				printf("Project successfully created: %s\n", pInReqInfo->szProject);
			}
			break;
		case E_FileOrProjectAlreadyExists:
			printf("Project not created:  Already Exists\n");
			break;
		default:
			printf("Error:  Server response = %d %s\n", 
				respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
	}

	return respMsg.eMsgNum;
}


/*----------------------------------------------
	Sends message to server to Add a file.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int AddFile(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	MessageInfo	respMsg;		/* Response Message to get from server */
	int nRetVal;				/* generic return value */

	/* Send the request message to the server.*/
	nRetVal = SendRequestAndGetResponse(nInSocket, E_AddFile,
				pInReqInfo, MSG_WITH_FILE, MSG_NO_FILE, &respMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}


	/* Print results */
	switch (respMsg.eMsgNum)
	{
		case E_Success :
		case E_FileAdded :
			{
				if (*ppRoot)
				{
					CLTreeLeaf* pLeaf = GetTreeLeafFromProject(*ppRoot, pInReqInfo->szProject, pInReqInfo->szFile);
					
					if (pLeaf)
					{
						STAT st;
						char szFileName[2048];
						
						memset(&st, 0, sizeof(STAT));
						
						if (pInReqInfo->szWorkdir != NULL)
							sprintf(szFileName, "%s%c%s", pInReqInfo->szWorkdir, SLASH, pInReqInfo->szFile);
						else
							strcpy(szFileName, pInReqInfo->szFile);
						
						if (!stat(szFileName, &st))
						{
							pLeaf->m_pLocalVersionDate = mktime(gmtime(&st.st_mtime));
							pLeaf->m_nLocalVersionNumber = GetIntHeaderValue(&respMsg.daHeaders, H_REVISION);
						}
					}
				}

				printf("File successfully added: %s\n", pInReqInfo->szFile);
			}
			break;
		case E_FileOrProjectAlreadyExists:
			printf("File not added:  Already Exists\n");
			break;
		default:
			printf("Error adding file: %s\tServer response = %d %s\n", pInReqInfo->szFile,
				respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
	}

	return respMsg.eMsgNum;
}

/*----------------------------------------------
	Sends message to server to Add a folder.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int AddFolder(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	int nRetVal = 0;				/* generic return value */
	char *pszFileName = NULL;
	char szFullName[2048];
	char szTemp[2048];
	char szTemp2[2048];
	char szFileSpec[2048];
#ifdef WIN32
	long Handle = 0;
#else
	void* Handle = NULL;
#endif

	RequestInfo TempRI;
	STAT st;

	memset(&st, 0, sizeof(STAT));

	if (stat(pInReqInfo->szWorkdir, &st))
	{
		printf("Could not open working directory: %s\n", pInReqInfo->szWorkdir);
		return -1;
	}

	memcpy(&TempRI, pInReqInfo, sizeof(RequestInfo));

	TempRI.eCommand = E_CreateProject;

	/* Add the initial project.*/
	if (CreateProject(nInSocket, ppRoot, &TempRI) < 0)
	{
		return -1;
	}


	sprintf(szFileSpec, "%s", pInReqInfo->szWorkdir);

	pszFileName = GetFirstFile(&Handle, szFileSpec);

	while (pszFileName != NULL)
	{
		if (!(!strcmp(pszFileName, ".") || !strcmp(pszFileName, "..")))
		{
			sprintf(szFullName, "%s%c%s", pInReqInfo->szWorkdir, SLASH, pszFileName);
			if (!stat(szFullName, &st))
			{
				if (st.st_mode & IFDIR)		/* the object is a directory */
				{
					if (pInReqInfo->bRecursive)
					{
						memcpy(&TempRI, pInReqInfo, sizeof(RequestInfo));
						
						sprintf(szTemp, "%s/%s", pInReqInfo->szProject, pszFileName);
						TempRI.szProject = szTemp;
						sprintf(szTemp2, "%s%c%s", pInReqInfo->szWorkdir, SLASH, pszFileName);
						TempRI.szWorkdir = szTemp2;
						TempRI.eCommand = E_AddFolder;
						
						nRetVal = AddFolder(nInSocket, ppRoot, &TempRI);
					}
				}
				else
				{
					memcpy(&TempRI, pInReqInfo, sizeof(RequestInfo));
					
					TempRI.szFile = pszFileName;
					TempRI.eCommand = E_AddFile;
					
					nRetVal = AddFile(nInSocket, ppRoot, &TempRI);
				}
			}
		}

		free(pszFileName);
		
		pszFileName = GetNextFile(&Handle);
	}
					
	return nRetVal;
}

/*----------------------------------------------
	Sends message to server to Get a file.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int GetFile(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	MessageInfo	respMsg;		/* Response Message to get from server */
	int nRetVal;				/* generic return value */


	if (CheckFileStatus(nInSocket, pInReqInfo, ppRoot, pInReqInfo->szProject, pInReqInfo->szFile, pInReqInfo->szWorkdir))
	{ 
		/* Send the request message to the server.*/
		nRetVal = SendRequestAndGetResponse(nInSocket, E_GetFile,
			pInReqInfo, MSG_NO_FILE, MSG_WITH_FILE, &respMsg);
		if (nRetVal < 0)
		{
			return nRetVal;
		}
		
		/* Print results */
		switch (respMsg.eMsgNum)
		{
		case E_File :
			{
				if (*ppRoot)
				{
					CLTreeLeaf* pLeaf = GetTreeLeafFromProject(*ppRoot, pInReqInfo->szProject, pInReqInfo->szFile);
					
					if (pLeaf)
					{
						STAT st;
						char szFileName[2048];
						
						memset(&st, 0, sizeof(STAT));
						
						if (pInReqInfo->szWorkdir != NULL)
							sprintf(szFileName, "%s%c%s", pInReqInfo->szWorkdir, SLASH, pInReqInfo->szFile);
						else
							strcpy(szFileName, pInReqInfo->szFile);
						
						if (!stat(szFileName, &st))
						{
							pLeaf->m_pLocalVersionDate = mktime(gmtime(&st.st_mtime));
							pLeaf->m_nLocalVersionNumber = GetIntHeaderValue(&respMsg.daHeaders, H_REVISION);
							pLeaf->m_lockedby = GetHeaderValue(&respMsg.daHeaders, H_LOCKED_BY);
						}
					}
				}
				
				if (pInReqInfo->bVerbose)
				{
					if (pInReqInfo->szWorkdir)
						printf("Got file: %s%c%s\n", pInReqInfo->szWorkdir, SLASH, pInReqInfo->szFile);
					else
						printf("Got file: %s\n", pInReqInfo->szFile);
				}
				/* Set permissions on file */
				ChangeDir(pInReqInfo->szWorkdir);
				if (pInReqInfo->bWriteable == 1)
					MakeFileReadWrite(pInReqInfo->szFile);
				else
					MakeFileReadOnly(pInReqInfo->szFile);
			}
			break;
		default:
			printf("Error:  Server response = %d %s\n", 
				respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
		}
		
		return respMsg.eMsgNum;
	} else {
		if (pInReqInfo->bVerbose)
			if (pInReqInfo->szWorkdir)
				printf("File already up to date: %s%c%s\n", pInReqInfo->szWorkdir, SLASH, pInReqInfo->szFile);
			else
				printf("File already up to date: %s\n", pInReqInfo->szFile);

		ChangeDir(pInReqInfo->szWorkdir);
		if (pInReqInfo->bWriteable == 1)
			MakeFileReadWrite(pInReqInfo->szFile);
		else
			MakeFileReadOnly(pInReqInfo->szFile);
		return E_File;
	}
}


/*----------------------------------------------
	Sends message to server to Undo a previous checkout
	of a file.  This unlocks the file
	and then replaces the local copy of it.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int UndoCheckOut(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	MessageInfo	respMsg;		/* Response Message to get from server */
	int nRetVal;				/* generic return value */

	/* Unlock the file.*/
	nRetVal = SendRequestAndGetResponse(nInSocket, E_UnlockFile,
				pInReqInfo, MSG_NO_FILE, MSG_NO_FILE, &respMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}

	/* Print results */
	switch (respMsg.eMsgNum)
	{
		/* If file was successfully unlocked, retrieve it again
		   so the user has the version under source safe, and not
		   the version on their hard disk */
		case E_Success :
		case E_FileUnlocked :
			printf("Unlocked file %s, now replacing local copy ...\n", pInReqInfo->szFile);
			nRetVal = GetFile(nInSocket, ppRoot, pInReqInfo);
			break;
		default:
			printf("Error:  Server response = %d %s\n", 
				respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
			nRetVal = respMsg.eMsgNum;
			break;
	}

	return nRetVal;
}

/*----------------------------------------------
	Un-Checks out all files in a project.
	
	Returns: server response code of last server
	response on success negative number on error.
----------------------------------------------*/
int UndoProjectCheckOut(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	int nRetVal = 0;
	RequestInfo tempRI;
	memcpy(&tempRI, pInReqInfo, sizeof(RequestInfo));
	
	tempRI.eCommand = E_GetProjectTree;
	
	nRetVal = GetInfo(nInSocket, ppRoot, &tempRI, E_ProjectTree, FALSE);
	
	if (pInReqInfo->szWorkdir == NULL)
	{
		pInReqInfo->szWorkdir = getcwd(pInReqInfo->szWorkdir, 1024);
		if (pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] == SLASH)
			pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] = 0;
	}

	ChangeDir(pInReqInfo->szWorkdir);
	
	nRetVal = GetProjectFiles(nInSocket, ppRoot, pInReqInfo, E_UndoCheckOut);
	return nRetVal;
}

/*----------------------------------------------
	Sends message to server to Checkout a file, which
	locks the file, then gets it.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int CheckOutFile(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	MessageInfo	respMsg;		/* Response Message to get from server */
	int nRetVal;				/* generic return value */
	CLTreeLeaf* pLeaf = NULL;

	nRetVal = GetFile(nInSocket, ppRoot, pInReqInfo);
	
	if (nRetVal == E_File)
	{
		if (pInReqInfo->szWorkdir != NULL)
		{
			if (chdir(pInReqInfo->szWorkdir) != 0)
				printf("Could not change to working directory %s.  Error Num=%d\n", pInReqInfo->szWorkdir, errno);
		}
		
		MakeFileReadWrite(pInReqInfo->szFile);

		if (*ppRoot)
		{
			pLeaf = GetTreeLeafFromProject(*ppRoot, pInReqInfo->szProject, pInReqInfo->szFile);
			if (BeginsWithIgnoreCase(pLeaf->m_lockedby, pInReqInfo->szName))
			{
				if (pInReqInfo->szWorkdir)
					printf("File already locked: %s%c%s\n", pInReqInfo->szWorkdir, SLASH, pInReqInfo->szFile);
				else
					printf("File already locked: %s\n", pInReqInfo->szFile);
				nRetVal = 304;
			}
			else
			{
				/* Lock the file.*/
				nRetVal = SendRequestAndGetResponse(nInSocket, E_LockFile,
					pInReqInfo, MSG_NO_FILE, MSG_NO_FILE, &respMsg);
				if (nRetVal < 0)
					return nRetVal;
				else
				{
					if (pInReqInfo->szWorkdir)
						printf("Checked out file: %s%c%s\n", pInReqInfo->szWorkdir, SLASH, pInReqInfo->szFile);
					else
						printf("Checked out file: %s\n", pInReqInfo->szFile);
					nRetVal = 304;
				}

				pLeaf->m_lockedby = pInReqInfo->szName;
			}
		}
		else
		{
			/* Lock the file.*/
			nRetVal = SendRequestAndGetResponse(nInSocket, E_LockFile,
				pInReqInfo, MSG_NO_FILE, MSG_NO_FILE, &respMsg);
			if (nRetVal < 0)
				return nRetVal;
			else
			{
				if (pInReqInfo->szWorkdir)
					printf("Checked out file: %s%c%s\n", pInReqInfo->szWorkdir, SLASH, pInReqInfo->szFile);
				else
					printf("Checked out file: %s\n", pInReqInfo->szFile);
				nRetVal = 304;
			}
		}
		
	} else {
		printf("Error:  Server response = %d %s\n", 
			nRetVal, ConvertMsgNum2MsgStr((enum MsgNum)nRetVal));
	}

	return nRetVal;
}

/*----------------------------------------------
	Checks out all files in a project.
	
	Returns: server response code of last server
	response on success negative number on error.
----------------------------------------------*/
int CheckOutProject(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	int nRetVal = 0;
	RequestInfo tempRI;
	memcpy(&tempRI, pInReqInfo, sizeof(RequestInfo));
	
	tempRI.eCommand = E_GetProjectTree;
	
	nRetVal = GetInfo(nInSocket, ppRoot, &tempRI, E_ProjectTree, FALSE);
	
	if (pInReqInfo->szWorkdir == NULL)
	{
		pInReqInfo->szWorkdir = getcwd(pInReqInfo->szWorkdir, 1024);
		if (pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] == SLASH)
			pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] = 0;
	}
	ChangeDir(pInReqInfo->szWorkdir);
	
	nRetVal = GetProjectFiles(nInSocket, ppRoot, pInReqInfo, E_CheckOutFile);
	return nRetVal;
}


/*----------------------------------------------
	Sends message to server to Checkin a file.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int CheckInFile(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	MessageInfo	respMsg;		/* Response Message to get from server */
	int nRetVal;				/* generic return value */


	/* Send the request message to the server.*/
	nRetVal = SendRequestAndGetResponse(nInSocket, E_CheckInFile,
				pInReqInfo, MSG_WITH_FILE, MSG_NO_FILE, &respMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}


	/* Print results */
	switch (respMsg.eMsgNum)
	{
		case E_Success :
		case E_CheckinComplete :
			{
				if (*ppRoot)
				{
					CLTreeLeaf* pLeaf = GetTreeLeafFromProject(*ppRoot, pInReqInfo->szProject, pInReqInfo->szFile);
					
					if (pLeaf)
					{
						STAT st;
						char szFileName[2048];
						
						memset(&st, 0, sizeof(STAT));
						
						if (pInReqInfo->szWorkdir != NULL)
							sprintf(szFileName, "%s%c%s", pInReqInfo->szWorkdir, SLASH, pInReqInfo->szFile);
						else
							strcpy(szFileName, pInReqInfo->szFile);
						
						if (!stat(szFileName, &st))
						{
							pLeaf->m_pLocalVersionDate = mktime(gmtime(&st.st_mtime));
							pLeaf->m_nLocalVersionNumber = GetIntHeaderValue(&respMsg.daHeaders, H_REVISION);
						}
					}
				}

				ChangeDir(pInReqInfo->szWorkdir);
				if (pInReqInfo->bWriteable == 1)
					MakeFileReadWrite(pInReqInfo->szFile);
				else
					MakeFileReadOnly(pInReqInfo->szFile);
				printf("File %s successfully checked in.\n", pInReqInfo->szFile);
			}
			break;
		case E_FileOrProjectNotLocked:
			printf("File Not Checked In:  You do not have this file checked out.\n");
			break;
		default:
			printf("Error:  Server response = %d %s\n", 
				respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
	}

	return respMsg.eMsgNum;
}

/*----------------------------------------------
	Checks in all files in a project.
	
	Returns: server response code of last server
	response on success negative number on error.
----------------------------------------------*/
int CheckInProject(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	int nRetVal = 0;
	RequestInfo tempRI;
	memcpy(&tempRI, pInReqInfo, sizeof(RequestInfo));
	
	tempRI.eCommand = E_GetProjectTree;
	
	nRetVal = GetInfo(nInSocket, ppRoot, &tempRI, E_ProjectTree, FALSE);
	
	if (pInReqInfo->szWorkdir == NULL)
	{
		pInReqInfo->szWorkdir = getcwd(pInReqInfo->szWorkdir, 1024);
		if (pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] == SLASH)
			pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] = 0;
	}
	ChangeDir(pInReqInfo->szWorkdir);
	
	nRetVal = GetProjectFiles(nInSocket, ppRoot, pInReqInfo, E_CheckInFile);
	return nRetVal;
}

/*----------------------------------------------
	Sends message to server to delete a file.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int DeleteFile(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	MessageInfo	respMsg;		/* Response Message to get from server */
	int nRetVal;				/* generic return value */


	/* Send the request message to the server.*/
	nRetVal = SendRequestAndGetResponse(nInSocket, E_Delete,
				pInReqInfo, MSG_NO_FILE, MSG_NO_FILE, &respMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}


	/* Print results */
	switch (respMsg.eMsgNum)
	{
		case E_Success :
			{
				if (*ppRoot)
				{
					CLTreeNode* pNode = GetTreeNodeFromProject(*ppRoot, pInReqInfo->szProject);
					if (pNode)
					{
						if (pInReqInfo->szFile == NULL)
						{
							TreeNodeDeleteChild(pNode->m_pParent, pNode);
						}
						else
						{
							CLTreeLeaf* pLeaf = GetTreeLeafFromProject(*ppRoot, pInReqInfo->szProject, pInReqInfo->szFile);
							
							if (pLeaf)
							{
								TreeNodeDeleteLeaf(pNode, pLeaf);
							}
						}
					}
				}

				if (pInReqInfo->szFile == NULL)
					printf("Project successfully deleted.\n");
				else
					printf("File successfully deleted.\n");
			}
			break;
		default:
			printf("Error:  Server response = %d %s\n", 
				respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
	}

	return respMsg.eMsgNum;
}

/*----------------------------------------------
	Sends message to server to rename a file.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int RenameFile(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	MessageInfo	respMsg;		/* Response Message to get from server */
	int nRetVal;				/* generic return value */


	/* Send the request message to the server.*/
	nRetVal = SendRequestAndGetResponse(nInSocket, E_Rename,
				pInReqInfo, MSG_NO_FILE, MSG_NO_FILE, &respMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}


	/* Print results */
	switch (respMsg.eMsgNum)
	{
		case E_Success :
			if (*ppRoot)
			{
				if (pInReqInfo->szFile)
				{
					CLTreeLeaf* pLeaf = GetTreeLeafFromProject(*ppRoot, pInReqInfo->szProject, pInReqInfo->szFile);
					
					if (pLeaf)
					{
						free(pLeaf->m_name);
						pLeaf->m_name = (char *)malloc(strlen(pInReqInfo->szNewName) + 1);
						if (pLeaf->m_name)
							memcpy(pLeaf->m_name, pInReqInfo->szNewName, strlen(pInReqInfo->szNewName) + 1);
					}
				}
				else
				{
					CLTreeNode* pNode = GetTreeNodeFromProject(*ppRoot, pInReqInfo->szProject);
					
					if (pNode)
					{
						free(pNode->m_name);
						pNode->m_name = (char *)malloc(strlen(pInReqInfo->szNewName) + 1);
						if (pNode->m_name)
							memcpy(pNode->m_name, pInReqInfo->szNewName, strlen(pInReqInfo->szNewName) + 1);
					}
				}
			}
			printf("File successfully renamed.\n");
			break;
		default:
			printf("Error:  Server response = %d %s\n", 
				respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
	}

	return respMsg.eMsgNum;
}


/*----------------------------------------------
	Sends message to server that retrieves information
	to be printed to the screen.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int GetInfo(
	int nInSocket,				/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,		/* Root of the SOS tree */
	RequestInfo *pInReqInfo,	/* Request struct containing commmand parameters */
	enum MsgNum	nInRespMsgNum,	/* The expected response msg num */
	short bPrintInfo			/* Print the info or not */
	)
{
	MessageInfo	respMsg;		/* Response Message to get from server */
	int nRetVal;				/* generic return value */
	CLTreeNode* pNode = NULL;

	/* Send the request message to the server.*/
	nRetVal = SendRequestAndGetResponse(nInSocket, pInReqInfo->eCommand,
				pInReqInfo, MSG_NO_FILE, MSG_NO_FILE, &respMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}

	if (ppRoot)
	{
		if (pInReqInfo->eCommand == E_GetProjectTree)
		{
			int pos = 0;
			int maxPos = respMsg.nBodySize - 1;
			short bParseError = FALSE;
			char* buf = NULL;
			char* nodeName = NULL;

			buf = (char *)malloc(respMsg.nBodySize + 1);
			memcpy(buf, respMsg.szMsgBody, respMsg.nBodySize + 1);

			nodeName = ParseNameFromBuffer(buf, &pos, maxPos, &bParseError);

			/* root node is null - downloading a new tree ... create a new root Node */
			if (!*ppRoot)
			{
				RequestInfo tempRI;
				memcpy(&tempRI, pInReqInfo, sizeof(RequestInfo));
		
				tempRI.eCommand = E_GetProjectTree;
				tempRI.szProject = ROOT_PROJECT;

/*				*ppRoot = TreeNodeWithName(nodeName);   */
				*ppRoot = TreeNodeWithName(ROOT_PROJECT);
	
				GetInfo(nInSocket, ppRoot, &tempRI, E_ProjectTree, FALSE);
			}

			/* pull out the node we are updating ... */
			pNode = GetTreeNodeFromProject(*ppRoot, pInReqInfo->szProject);

			if (pNode)
			{
				UpdateProjectTree(buf, pNode, &pos, maxPos);
			}
		}
		else if (pInReqInfo->eCommand == E_GetFileInfo)
		{
			if (*ppRoot)
			{
				CLTreeLeaf* pLeaf = GetTreeLeafFromProject(*ppRoot, pInReqInfo->szProject, pInReqInfo->szFile);
				char* header;

				if (!pLeaf)
					return -1;

				/* update the leaf */
				header = GetHeaderValue(&respMsg.daHeaders, H_DATE);

				if (header)
					pLeaf->m_pLocalVersionDate = ParseServerDate(header);

				header = GetHeaderValue(&respMsg.daHeaders, H_REVISION);

				if (header)
					pLeaf->m_nRemoteVersionNumber = atoi(header);

			}
		}
		else if (pInReqInfo->eCommand == E_GetFileList)
		{
			if (*ppRoot)
			{
				pNode = GetTreeNodeFromProject(*ppRoot, pInReqInfo->szProject);
				ParseFileList(pNode, respMsg.szMsgBody, respMsg.nBodySize); 
			}
		}
	}
	

	/* Print results */
	if (respMsg.eMsgNum == nInRespMsgNum)
	{
		if (bPrintInfo == TRUE)
		{
			switch (nInRespMsgNum)
			{
			case E_ProjectTree:
				{
					if (pNode)
						PrintTreeNode(pNode, 0);
					break;
				}
			default:
				{
					/* Print relevant headers */
					char *header;
					header = GetHeaderValue(&respMsg.daHeaders, H_REVISION);
					if (header != NULL)
						printf("Version: %s\n", header);
					
					header = GetHeaderValue(&respMsg.daHeaders, H_DATE);
					if (header != NULL)
						printf("Remote Date: %s\n", header);
					
					header = GetHeaderValue(&respMsg.daHeaders, H_LOCKED_BY);
					if (header != NULL)
						printf("Locked By: %s\n", header);
					
					printf("%s\n", respMsg.szMsgBody);
				}
			}
		}
	} else {
		printf("Error:  Server response = %d %s\n", 
			respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
	}

	return respMsg.eMsgNum;
}

/*----------------------------------------------
	Determine the working directory to place a file returned from
	a GetProject, based on the original project retrieved.  If
	the file's project is a subdirectory off the original project,
	then cd to a working directory in the same place relative to
	the original working directory.

	For example, If orig project is /dogs/cats, and the current
	project is /dogs/cats/bears/weasel, then we need to cd to the
	directory bears/weasel relative to the original workdir.

	This function creates any directories necessary to move
	to the computed working directory


	Returns: nothing
----------------------------------------------*/
void GoToWorkDir(
	char *szInOrigProject,	/* the original project being retrieved */
	char *szInCurProject,	/* the project of the current file being retrieved */
	char *szInOrigWorkDir	/* the original working dir to place files */
	)
{
	int idx;		/* array index */

	/* First, change to the original work dir */
	ChangeDir(szInOrigWorkDir);

	/* If the projects are not the same */
	if (strcmp(szInOrigProject, szInCurProject))
	{

		/* 
			Get the extra part of the cur project 
		*/
		/* screw case - if the cur project is less than orig proj, exit */
		if (strlen(szInOrigProject) > strlen(szInCurProject))
		{
			return;
		}

		
		idx = strlen(szInOrigProject);
		while (szInCurProject[idx] == '/' || 
			   szInCurProject[idx] == '\\')
		{
			idx++;
		}

		/* cd to the new work dir, creating intermediate dirs as necessary */
		ChangeDir(szInCurProject + idx);

	}
	
}

/*----------------------------------------------
	Sends message to server that retrieves all files
	in a project, and optionally all files in all
	subprojects of the project (recursively).

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int GetProject(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo /* Request struct containing commmand parameters */
	)
{

	int nRetVal;				/* generic return value */
	MessageInfo	respMsg;		/* Response Message to get from server */

	if (pInReqInfo->szLabel != NULL)
	{
		pInReqInfo->bAlwaysDownload = 1;
	}
	
	if (pInReqInfo->bAlwaysDownload == 1)
	{
		char *pTempFileName;		/* Name of temporary file */
		char szOrigWorkDir[1024];	/* full path name of original working dir */


		/* First, get the full path name of the working directory if it
		exists.  This is needed to reset to the initial working dir
		when files are retrieved recursively */
		if (pInReqInfo->szWorkdir != NULL)
		{
			ChangeDir(pInReqInfo->szWorkdir);
		}
		getcwd(szOrigWorkDir, 1024);
		
		
		/* Ensure the response message is initialized */
		MI_Initialize(&respMsg);
		
		/* Create the server request message */
		CreateServerMessage(pInReqInfo->eCommand, pInReqInfo, &respMsg);
		
		nRetVal = Net_SendRequest(nInSocket, &respMsg, pInReqInfo);
		
		if (nRetVal < 0)
		{
			return nRetVal;
		}
		
		/* 
		The response from the server should be a series of 203 File messages
		followed by a single 300 Success.  The returning response should
		contain the name of the file, so initially place the file in a
		temporary place, then move to the final location.  If any response other
		than a 203 is returned, stop retrieving files from the server.  If
		its not a 300 Success, also output an error message
		*/
		respMsg.eMsgNum = E_File;
		while (nRetVal >= 0 && respMsg.eMsgNum == E_File)
		{
			/* Get the file */
			pTempFileName = tmpnam(NULL);
			nRetVal = Net_GetResponse(nInSocket, &respMsg, pInReqInfo,
				pTempFileName, NULL, 1);
			
			if (respMsg.eMsgNum != E_File)
				break;
			
			/* Change to the correct working directory */
			GoToWorkDir(pInReqInfo->szProject, 
				GetHeaderValue(&respMsg.daHeaders, H_CUR_PROJECT),
				szOrigWorkDir);
			
			/* Output the name of the file if verbose */
			if (pInReqInfo->bVerbose)
			{
				char buf[1024];
				getcwd(buf, 1024);
				printf("Got file: %s%c%s\n", buf, SLASH, 
					GetHeaderValue(&respMsg.daHeaders, H_FILE_NAME));
			}
			
			/* Move the file to the working directory */
			MoveFileToCurDir(pTempFileName, GetHeaderValue(&respMsg.daHeaders, H_FILE_NAME));
			
			if (pInReqInfo->bWriteable == 0)
				MakeFileReadOnly(GetHeaderValue(&respMsg.daHeaders, H_FILE_NAME));
			
			{
				char *pProject = GetHeaderValue(&respMsg.daHeaders, H_CUR_PROJECT);
				char *pFile = GetHeaderValue(&respMsg.daHeaders, H_FILE_NAME);

				if (*ppRoot)
				{
					CLTreeLeaf* pLeaf = GetTreeLeafFromProject(*ppRoot, pProject, pFile);
					
					if (pLeaf)
					{
						STAT st;
						
						memset(&st, 0, sizeof(STAT));
						
						if (!stat(pFile, &st))
						{
							pLeaf->m_pLocalVersionDate = mktime(gmtime(&st.st_mtime));
							pLeaf->m_nLocalVersionNumber = GetIntHeaderValue(&respMsg.daHeaders, H_REVISION);
						}
					}
				}
			}
			
			
		}
		
		
		/* Print results */
		if (respMsg.eMsgNum == E_Success)
		{
			printf("Finished retrieving project files.\n");
		} else {
			printf("Error:  Server response = %d %s\n", 
				respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
		}
		
		return respMsg.eMsgNum;
	}
	else
	{
		RequestInfo tempRI;
		memcpy(&tempRI, pInReqInfo, sizeof(RequestInfo));

		tempRI.eCommand = E_GetProjectTree;

		nRetVal = GetInfo(nInSocket, ppRoot, &tempRI, E_ProjectTree, FALSE);

		if (pInReqInfo->szWorkdir == NULL)
		{
			pInReqInfo->szWorkdir = getcwd(pInReqInfo->szWorkdir, 1024);
			if (pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] == SLASH)
				pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] = 0;
		}
		ChangeDir(pInReqInfo->szWorkdir);

		nRetVal = GetProjectFiles(nInSocket, ppRoot, pInReqInfo, E_GetFile);
		return nRetVal;
	}
}




/*----------------------------------------------
Steps through all the files in a given project 
and performs a specified action upon them.
If -recursive is specified, will recurse
down through the project tree.

Returns: server response code of last server
response on success negative number on error.
  ----------------------------------------------*/
int GetProjectFiles(int nInSocket, CLTreeNode** ppRoot, RequestInfo* pInReqInfo, enum MsgNum eAction)
{
	int nRetVal = 0;
	CLTreeNode* pNode = NULL;
	CLTreeLeafList* pLeaves = NULL;
	CLTreeNodeList* pNodes = NULL;
	RequestInfo tempRI;

	memcpy(&tempRI, pInReqInfo, sizeof(RequestInfo));
	tempRI.eCommand = E_GetFileList;

	nRetVal = GetInfo(nInSocket, ppRoot, &tempRI, E_FileList, FALSE);

	pNode = GetTreeNodeFromProject(*ppRoot, pInReqInfo->szProject);
	if (!pNode)
		return -1;

	pLeaves = pNode->m_pFiles;

	while (pLeaves)
	{
		memcpy(&tempRI, pInReqInfo, sizeof(RequestInfo));

		tempRI.eCommand = eAction;
		tempRI.szFile = pLeaves->m_pLeaf->m_name;

		switch (eAction)
		{
		case E_GetFile:
			nRetVal = GetFile(nInSocket, ppRoot, &tempRI);
			break;
		case E_CheckOutFile:
			nRetVal = CheckOutFile(nInSocket, ppRoot, &tempRI);
			break;
		case E_CheckInFile:
			nRetVal = CheckInFile(nInSocket, ppRoot, &tempRI);
			break;
		case E_UndoCheckOut:
			nRetVal = UndoCheckOut(nInSocket, ppRoot, &tempRI);
			break;
		case E_ShareFile:
			nRetVal = ShareFile(nInSocket, ppRoot, &tempRI);
			break;
		case E_BranchFile:
			nRetVal = BranchFile(nInSocket, ppRoot, &tempRI);
			break;
		default:
			return -1;
		}

		pLeaves = pLeaves->m_pNext;
	}

	if (pInReqInfo->bRecursive == 1)
	{
		pNodes = pNode->m_pChildren;

		while (pNodes)
		{
			char szProjName[2048];
			char szWorkDir[2048];
			char szShareTo[2048];

			memcpy(&tempRI, pInReqInfo, sizeof(RequestInfo));
			
			if (tempRI.szProject[strlen(tempRI.szProject) - 1] == '/')
				sprintf(szProjName, "%s%s", tempRI.szProject, pNodes->m_pNode->m_name);
			else
				sprintf(szProjName, "%s/%s", tempRI.szProject, pNodes->m_pNode->m_name);
			tempRI.szProject = szProjName;

			if (tempRI.eCommand == E_ShareProject)
			{
				if (tempRI.szShareTo[strlen(tempRI.szShareTo) - 1] == '/')
					sprintf(szShareTo, "%s%s", tempRI.szShareTo, pNodes->m_pNode->m_name);
				else
					sprintf(szShareTo, "%s/%s", tempRI.szShareTo, pNodes->m_pNode->m_name);
				tempRI.szShareTo = szShareTo;
			}

			sprintf(szWorkDir, "%s%c%s", tempRI.szWorkdir, SLASH, pNodes->m_pNode->m_name);
			tempRI.szWorkdir = szWorkDir;

			ChangeDir(tempRI.szWorkdir);
			
			nRetVal = GetProjectFiles(nInSocket, ppRoot, &tempRI, eAction);
			
			pNodes = pNodes->m_pNext;
		}
	}

	return nRetVal;

}


/*----------------------------------------------
	Shares a file into a new project.  Optionally
	can branch the file after sharing it if the 
	-branch parameter is specified.

	Returns: server response code of last server
	response on success or negative number on error.
----------------------------------------------*/

int ShareFile(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pReqInfo /* Request struct containing commmand parameters */
	)
{

	MessageInfo	RespMsg;
	int nRetVal = 0;

	SendRequestAndGetResponse(nInSocket, E_ShareFile, pReqInfo, MSG_NO_FILE, MSG_WITH_FILE, &RespMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}

	switch (RespMsg.eMsgNum)
	{
		case E_File :
			{
				if (*ppRoot)
				{
					CLTreeLeaf* pLeaf = GetTreeLeafFromProject(*ppRoot, pReqInfo->szShareTo, pReqInfo->szFile);
					
					if (pLeaf)
					{
						STAT st;
						char szFileName[2048];
						
						memset(&st, 0, sizeof(STAT));
						
						if (pReqInfo->szWorkdir != NULL)
							sprintf(szFileName, "%s%c%s", pReqInfo->szWorkdir, SLASH, pReqInfo->szFile);
						else
							strcpy(szFileName, pReqInfo->szFile);
						
						if (!stat(szFileName, &st))
						{
							pLeaf->m_pLocalVersionDate = mktime(gmtime(&st.st_mtime));
							pLeaf->m_nLocalVersionNumber = GetIntHeaderValue(&RespMsg.daHeaders, H_REVISION);
						}
					}
				}
				
				if (pReqInfo->bBranchAfterShare == 1)
					printf("Shared and branched file %s\n", pReqInfo->szFile);
				else
					printf("Shared file %s\n", pReqInfo->szFile);

				/* Set permissions on file to read-only */
				if (pReqInfo->bWriteable == 0)
				{
					ChangeDir(pReqInfo->szWorkdir);
					MakeFileReadOnly(pReqInfo->szFile);
				}
			}
			break;
		default:
			printf("Error:  Server response = %d %s\n", 
				RespMsg.eMsgNum, ConvertMsgNum2MsgStr(RespMsg.eMsgNum));
	}


	return RespMsg.eMsgNum;

}

/*----------------------------------------------
	Shares all files in a project.
	
	Returns: server response code of last server
	response on success negative number on error.
----------------------------------------------*/
int ShareProject(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	int nRetVal;				/* generic return value */
	MessageInfo	respMsg;		/* Response Message to get from server */
	char *pTempFileName;		/* Name of temporary file */
	char szOrigWorkDir[1024];	/* full path name of original working dir */
	
	
	/* First, get the full path name of the working directory if it
	exists.  This is needed to reset to the initial working dir
	when files are retrieved recursively */
	if (pInReqInfo->szWorkdir != NULL)
	{
		ChangeDir(pInReqInfo->szWorkdir);
	}
	getcwd(szOrigWorkDir, 1024);
	
	
	/* Ensure the response message is initialized */
	MI_Initialize(&respMsg);
	
	/* Create the server request message */
	CreateServerMessage(pInReqInfo->eCommand, pInReqInfo, &respMsg);
	
	nRetVal = Net_SendRequest(nInSocket, &respMsg, pInReqInfo);
	
	if (nRetVal < 0)
	{
		return nRetVal;
	}
	
	/* 
	The response from the server should be a series of 203 File messages
	followed by a single 300 Success.  The returning response should
	contain the name of the file, so initially place the file in a
	temporary place, then move to the final location.  If any response other
	than a 203 is returned, stop retrieving files from the server.  If
	its not a 300 Success, also output an error message
	*/
	respMsg.eMsgNum = E_File;
	while (nRetVal >= 0 && respMsg.eMsgNum == E_File)
	{
		/* Get the file */
		pTempFileName = tmpnam(NULL);
		nRetVal = Net_GetResponse(nInSocket, &respMsg, pInReqInfo,
			pTempFileName, NULL, 1);
		
		if (respMsg.eMsgNum != E_File)
			break;
		
		/* Change to the correct working directory */
		GoToWorkDir(pInReqInfo->szProject, 
			GetHeaderValue(&respMsg.daHeaders, H_CUR_PROJECT),
			szOrigWorkDir);
		
		/* Output the name of the file if verbose */
		if (pInReqInfo->bVerbose)
		{
			char buf[1024];
			getcwd(buf, 1024);
			printf("Shared file: %s%c%s\n", buf, SLASH, 
				GetHeaderValue(&respMsg.daHeaders, H_FILE_NAME));
		}
		
		/* Move the file to the working directory */
		MoveFileToCurDir(pTempFileName, GetHeaderValue(&respMsg.daHeaders, H_FILE_NAME));
		
		if (pInReqInfo->bWriteable == 0)
			MakeFileReadOnly(GetHeaderValue(&respMsg.daHeaders, H_FILE_NAME));
		
		{
			char *pProject = GetHeaderValue(&respMsg.daHeaders, H_CUR_PROJECT);
			char *pFile = GetHeaderValue(&respMsg.daHeaders, H_FILE_NAME);
			
			if (*ppRoot)
			{
				CLTreeLeaf* pLeaf = GetTreeLeafFromProject(*ppRoot, pProject, pFile);
				
				if (pLeaf)
				{
					STAT st;
					
					memset(&st, 0, sizeof(STAT));
					
					if (!stat(pFile, &st))
					{
						pLeaf->m_pLocalVersionDate = mktime(gmtime(&st.st_mtime));
						pLeaf->m_nLocalVersionNumber = GetIntHeaderValue(&respMsg.daHeaders, H_REVISION);
					}
				}
			}
		}
		
		
	}
	
	
	/* Print results */
	if (respMsg.eMsgNum == E_Success)
	{
		if (pInReqInfo->bBranchAfterShare == 1)
			printf("Finished sharing and branching project files.\n");
		else
			printf("Finished sharing project files.\n");
	} else {
		printf("Error:  Server response = %d %s\n", 
			respMsg.eMsgNum, ConvertMsgNum2MsgStr(respMsg.eMsgNum));
	}
	
	return respMsg.eMsgNum;

}

/*----------------------------------------------
	Branches a file that was previously shared.

	Returns: server response code of last server
	response on success or negative number on error.
----------------------------------------------*/

int BranchFile(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pReqInfo /* Request struct containing commmand parameters */
	)
{

	MessageInfo	RespMsg;
	int nRetVal = 0;

	SendRequestAndGetResponse(nInSocket, E_BranchFile, pReqInfo, MSG_NO_FILE, MSG_NO_FILE, &RespMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}

	switch (RespMsg.eMsgNum)
	{
		case E_Success :
			printf("Branched file %s\n", pReqInfo->szFile);
			break;
		default:
			printf("Error:  Server response = %d %s\n", 
				RespMsg.eMsgNum, ConvertMsgNum2MsgStr(RespMsg.eMsgNum));
	}


	return RespMsg.eMsgNum;

}


/*----------------------------------------------
	Branches all files in a project.
	
	Returns: server response code of last server
	response on success negative number on error.
----------------------------------------------*/
int BranchProject(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pInReqInfo	/* Request struct containing commmand parameters */
	)
{
	int nRetVal = 0;
	RequestInfo tempRI;
	memcpy(&tempRI, pInReqInfo, sizeof(RequestInfo));
	
	tempRI.eCommand = E_GetProjectTree;
	
	nRetVal = GetInfo(nInSocket, ppRoot, &tempRI, E_ProjectTree, FALSE);
	
	if (pInReqInfo->szWorkdir == NULL)
	{
		pInReqInfo->szWorkdir = getcwd(pInReqInfo->szWorkdir, 1024);
		if (pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] == SLASH)
			pInReqInfo->szWorkdir[strlen(pInReqInfo->szWorkdir) - 1] = 0;
	}
	ChangeDir(pInReqInfo->szWorkdir);
	
	nRetVal = GetProjectFiles(nInSocket, ppRoot, pInReqInfo, E_BranchFile);
	return nRetVal;
}

/*----------------------------------------------
	Adds a label to a project or file.

	Returns: server response code of last server
	response on success or negative number on error.
----------------------------------------------*/

int AddLabel(
	int nInSocket,			/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,	/* SOS Tree */
	RequestInfo *pReqInfo /* Request struct containing commmand parameters */
	)
{

	MessageInfo	RespMsg;
	int nRetVal = 0;

	SendRequestAndGetResponse(nInSocket, E_AddLabel, pReqInfo, MSG_NO_FILE, MSG_NO_FILE, &RespMsg);
	if (nRetVal < 0)
	{
		return nRetVal;
	}

	switch (RespMsg.eMsgNum)
	{
		case E_Success :
			if (pReqInfo->szFile)
			{
				RequestInfo rInfo;
				
				memcpy(&rInfo, pReqInfo, sizeof(RequestInfo));
				
				rInfo.eCommand = E_GetFileInfo;

				nRetVal = GetInfo(nInSocket, ppRoot, &rInfo, E_FileInfo, FALSE);
			}

			printf("Added label %s to %s/%s\n", pReqInfo->szLabel, pReqInfo->szProject, pReqInfo->szFile?pReqInfo->szFile:" ");
			break;
		default:
			printf("Error:  Server response = %d %s\n", 
				RespMsg.eMsgNum, ConvertMsgNum2MsgStr(RespMsg.eMsgNum));
	}


	return RespMsg.eMsgNum;

}


/*----------------------------------------------
	Outputs the list of databases returned by the
	server in the expected Challenge response.  This
	function should only be called once, after 
	the intial connection to the server.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int GetDatabases(
	int nInSocket,			/* an open socket connected to a SS Server */
	RequestInfo *pReqInfo
	)
{

	MessageInfo	serverMsg;		/* server response message */
	int nRetVal;				/* generic return value */

#ifndef BUILD_STANDARD
	char *header;
	int nHeadLen;
#endif /* !BUILD_STANDARD */

	MI_Initialize(&serverMsg);

	/* Get the expected challenge as the first response
		message by the server
	*/
	nRetVal = Net_GetEntireResponse(nInSocket,&serverMsg, pReqInfo);
	if (nRetVal < 0)
	{
		return nRetVal;
	}

	/* Print results */
	switch (serverMsg.eMsgNum)
	{
		case E_Challenge :
			printf("%s\n", serverMsg.szMsgBody);
			break;
#ifndef BUILD_STANDARD
		case E_SecureChallenge :
			header = GetHeaderValue(&serverMsg.daHeaders, H_KEY_LEN);
			nHeadLen = atoi(header);
			printf("%s\n", serverMsg.szMsgBody + nHeadLen);
			break;
#endif /* !BUILD_STANDARD */
		default:
			printf("Error:  Server response = %d %s\n", 
				serverMsg.eMsgNum, ConvertMsgNum2MsgStr(serverMsg.eMsgNum));
	}

	return serverMsg.eMsgNum;
}


/*----------------------------------------------
Checks a file's date and version in the SOS file 
against the version on the server.  

Returns a 1 if the file needs to be fetched and 
a 0 if the file is already up to date.
----------------------------------------------*/
int CheckFileStatus(int nSocket,
					RequestInfo* pInReqInfo,
					CLTreeNode** ppRoot, 
					char* pszProject, 
					char* pszFile, 
					char* pszWorkDir)
{
	CLTreeLeaf* pLeaf = NULL;
	STAT st;
	char szFileName[2048];
	int nRetVal = 0;
	RequestInfo ReqInfo;
	MessageInfo RespMsg;

	if (pInReqInfo->bAlwaysDownload == 1)
		return 1;

	if (!*ppRoot)
		return 1;

	pLeaf = GetTreeLeafFromProject(*ppRoot, pszProject, pszFile);

	if (!pLeaf)
		return 1;

	if ((pLeaf->m_nLocalVersionNumber == -1) || (pLeaf->m_pLocalVersionDate == 0))
		return 1;

	memset(&st, 0, sizeof(STAT));

	if (pszWorkDir)
		sprintf(szFileName, "%s%c%s", pszWorkDir, SLASH, pszFile);
	else
		sprintf(szFileName, "%s", pszFile);

	if (!stat(szFileName, &st))
	{
		if (pLeaf->m_pLocalVersionDate != mktime(gmtime(&st.st_mtime)))
			return 1;

		memcpy(&ReqInfo, pInReqInfo, sizeof(RequestInfo));
		ReqInfo.szProject = pszProject;
		ReqInfo.szFile = pszFile;

		nRetVal = SendRequestAndGetResponse(nSocket, E_GetFileInfo,
			&ReqInfo, MSG_NO_FILE, MSG_NO_FILE, &RespMsg);

		if (pLeaf->m_lockedby == NULL)
			pLeaf->m_lockedby = GetHeaderValue(&RespMsg.daHeaders, H_LOCKED_BY);

		if (GetIntHeaderValue(&RespMsg.daHeaders, H_REVISION) != pLeaf->m_nLocalVersionNumber)
			return 1;
	} else {
		return 1;
	}

	return 0;
}


/*----------------------------------------------
	Route the input command to the procedure
	that implements the command.

	Assumes the command has been verified to have 
	correct parameters.

	Returns: server response code of last server
	response on success.
	negative number on error.
----------------------------------------------*/
int ExecuteCommand(
	int nSocket,				/* an open socket connected to a SS Server */
	CLTreeNode** ppRoot,		/* Root of the SOS Tree */
	RequestInfo *pReqInfo,		/* Request struct containing commmand */
								/* and the command parameters */
	short bLoadTreeFromServer	/* should the tree be downloaded from the server? */
	)
{
	int nRetVal = -1;

	/* we have a SOS file - use it to get info for the command being run */
	if (*ppRoot)
	{
		/* We've been told to use a SOS file, but there's no tree - download it .. */
		if (bLoadTreeFromServer)
		{
			RequestInfo rInfo;

			memcpy(&rInfo, pReqInfo, sizeof(RequestInfo));

			rInfo.szProject = ROOT_PROJECT;
			rInfo.eCommand = E_GetProjectTree;

			if (rInfo.bVerbose)
				printf("Downloading project tree from server\n");

			nRetVal = GetInfo(nSocket, ppRoot, &rInfo, E_ProjectTree, FALSE);
		}

	}

	switch(pReqInfo->eCommand)
	{
		case E_CreateProject:
			nRetVal = CreateProject(nSocket, ppRoot, pReqInfo);
			break;
		case E_AddFile:
			nRetVal = AddFile(nSocket, ppRoot, pReqInfo);
			break;
		case E_GetFile:
			nRetVal = GetFile(nSocket, ppRoot, pReqInfo);
			break;
		case E_CheckOutFile:
			nRetVal = CheckOutFile(nSocket, ppRoot, pReqInfo);
			break;
		case E_CheckOutProject:
			nRetVal = CheckOutProject(nSocket, ppRoot, pReqInfo);
			break;
		case E_UndoCheckOut:
			nRetVal = UndoCheckOut(nSocket, ppRoot, pReqInfo);
			break;
		case E_UndoProjectCheckOut:
			nRetVal = UndoProjectCheckOut(nSocket, ppRoot, pReqInfo);
			break;
		case E_CheckInFile:
			nRetVal = CheckInFile(nSocket, ppRoot, pReqInfo);
			break;
		case E_CheckInProject:
			nRetVal = CheckInProject(nSocket, ppRoot, pReqInfo);
			break;
		case E_Delete:
			nRetVal = DeleteFile(nSocket, ppRoot, pReqInfo);
			break;
		case E_Rename:
			nRetVal = RenameFile(nSocket, ppRoot, pReqInfo);
			break;
		case E_GetProjectTree:
			if (pReqInfo->szProject == NULL)		/* Default to root project */
				pReqInfo->szProject = ROOT_PROJECT;
			nRetVal = GetInfo(nSocket, ppRoot, pReqInfo, E_ProjectTree, TRUE);
			break;
		case E_GetFileList:
			nRetVal = GetInfo(nSocket, ppRoot, pReqInfo, E_FileList, TRUE);
			break;
		case E_GetFileHistory:
			nRetVal = GetInfo(nSocket, ppRoot, pReqInfo, E_FileHistory, TRUE);
			break;
		case E_GetFileInfo:
			nRetVal = GetInfo(nSocket, ppRoot, pReqInfo, E_FileInfo, TRUE);
			break;

		case E_GetProject:
			nRetVal = GetProject(nSocket, ppRoot, pReqInfo);
			break;
		case E_ShareFile:
			nRetVal = ShareFile(nSocket, ppRoot, pReqInfo);
			break;
		case E_ShareProject:
			nRetVal = ShareProject(nSocket, ppRoot, pReqInfo);
			break;
		case E_BranchFile:
			nRetVal = BranchFile(nSocket, ppRoot, pReqInfo);
			break;
		case E_BranchProject:
			nRetVal = BranchProject(nSocket, ppRoot, pReqInfo);
			break;
		case E_AddLabel:
			nRetVal = AddLabel(nSocket, ppRoot, pReqInfo);
			break;
		case E_AddFolder:
			nRetVal = AddFolder(nSocket, ppRoot, pReqInfo);
			break;
		default:
			printf("Command %d (\"%s\") not recognized as valid SOS command\n", 
				pReqInfo->eCommand, pReqInfo->szOrigInpCmd);

	}

	return nRetVal;
}



