
/*******************************************************

	FILE: CLcommon.c

	DESC: This file contains definitions for functions 
		common to all parts of the command line client.
 
        Copyright 1998-2002 SourceGear Corporation.
        All Rights Reserved. 

*******************************************************/

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <ctype.h>

#ifdef WIN32
#include <io.h>
#include <direct.h>
#endif /* WIN32 */

#ifdef UNIX
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#endif /* UNIX */

#include "CLcommon.h"
#include "CLTree.h"


/*--- D Y N A M I C    A R R A Y    F U N C T I O N S ------*/

/*----------------------------------------------
	Initialize the contents of a dynamic array.
	MUST BE CALLED AFTER BEFORE USING OTHER DARRAY FUNCTIONS

	Returns: nothing
----------------------------------------------*/
void DA_Initialize(DArray *theArray)
{
	theArray->szArray = NULL;
	theArray->nCapacity=0;
	theArray->nCurLen=0;
}
  
/*----------------------------------------------
	Free any allocated memory in a dynamic array
	and reset the internal state.

	Returns: nothing
----------------------------------------------*/
void DA_Destroy(DArray *theArray)
{
	DA_Reset(theArray);
}

/*----------------------------------------------
	Free any allocated memory in a dynamic array
	and reset the internal state.

	Returns: nothing
----------------------------------------------*/
void DA_Reset(DArray *theArray)
{
	if (theArray->szArray != NULL)
	{
		free(theArray->szArray);
		theArray->szArray = NULL;
	}
	theArray->nCapacity=0;
	theArray->nCurLen=0;
}

/*----------------------------------------------
	Ensure that the input DArray has enough capacity
	to assign the input number of bytes.  If it doesn't,
	it grows the array so that it does have enough capacity.

	Returns: 0 if there is not enough memory to grow the array.
			 1 for a success.
----------------------------------------------*/
int DA_EnsureCapacity(
	DArray *theArray,	/* the array to ensure capacity of */
	int nBufSize		/* the number of bytes required	*/
	)
{
	char *pNewArray;
	int nGrowSize=0;

	/*  Determine whether the DArray needs to grow to accomodate
		the input string, and if so, by how much */
	while (nBufSize > nGrowSize + theArray->nCapacity)
	{
		nGrowSize += DARRAY_GROW_INCR;
	}

	/*  If the array is too small, then grow the array by the 
		previously computed amount */
	if (nGrowSize > 0)
	{
		pNewArray = (char *)malloc(nGrowSize + theArray->nCapacity + 1);
		if (theArray->szArray != NULL)
		{
			memcpy(pNewArray, theArray->szArray, theArray->nCurLen);
			free(theArray->szArray);
		}
		theArray->szArray = pNewArray;
		theArray->nCapacity += nGrowSize;
	}

	/* If the malloc failed, return error */
	if (theArray->szArray == NULL)
	{
		return 0;
	}

	return 1;
}

/*----------------------------------------------
	Assign a character string to a dynamic array, 
	resizing if necessary.  Null terminates the 
	string on assignment.

	Returns: 0 if there is not enough memory to grow the array.
			 1 for a successful assignment.
			 negative number on error
----------------------------------------------*/
int DA_Assign(
	DArray *theArray,	/* the array to assign the string to */
	char *szBuf,		/* the string to copy to the DArray */
	int nBufSize		/* the number of bytes to copy.  If < 0, the input	*/
						/* string is assumed to be zero-terminated, and		*/
						/* the entire input string is copied to the DArray	*/
	)
{
	if (theArray == 0 || szBuf == 0)
	{
		return -1;
	}

	/*  if the buf size is less than zero, assume the input
		string is 0 terminated, so get its length */
	if (nBufSize < 0)
	{
		nBufSize = strlen(szBuf);
	}

	if (DA_EnsureCapacity(theArray, nBufSize) == 0)
	{
		return 0;
	}

	/* Copy the string to the array, and null terminate it */
	memcpy(theArray->szArray, szBuf, nBufSize);
	theArray->nCurLen = nBufSize;
	theArray->szArray[nBufSize] = '\0';

	return 1;
}

/*--------------------------------------------------------
	Append a character string to a dynamic array, resizing
	if necessary.  Null terminates the string after appending.

  Returns: 0 if there is not enough memory to grow the array.
			 1 for a successful append.
			 negative number on error
----------------------------------------------------------*/
int DA_Append(
	DArray *theArray,	/* the array to append the string to */
	char *szBuf,		/* the string to append to the DArray */
	int nBufSize		/* the number of bytes to copy.  If < 0, the input	*/
						/* string is assumed to be zero-terminated, and		*/
						/* the entire input string is copied to the DArray	*/
	)
{
	if (theArray == 0 || szBuf == 0)
	{
		return -1;
	}

	/*	If the DArray is currently empty, just use the assign function */
	if (theArray->nCurLen < 1)
	{
		return DA_Assign(theArray, szBuf, nBufSize);
	}

	/*  If the buf size is less than zero, assume the input
		string is 0 terminated, so get its length */
	if (nBufSize < 0)
	{
		nBufSize = strlen(szBuf);
	}


	if (DA_EnsureCapacity(theArray, nBufSize + theArray->nCurLen) == 0)
	{
		return 0;
	}

	/* Append the string to the array, and null terminate it */
	memcpy(theArray->szArray + theArray->nCurLen,
				szBuf, nBufSize);
	theArray->nCurLen += nBufSize;
	theArray->szArray[theArray->nCurLen] = '\0';

	return 1;

}

/*----------------------------------------------
	Get the character in a dynamic array
	at the position specified by the input index.

	Returns: The char at the specified index or the null char 
			if the input index is out of range
----------------------------------------------*/
char DA_GetChar(
	DArray *theArray,	/* the array to get the char from */
	int nIdx			/* the index of the array to get the char from */
	)
{
	if (nIdx < 0 || nIdx > theArray->nCurLen)
		return '\0';
	else
		return theArray->szArray[nIdx];

}

/*----------------------------------------------
	Set the character in a dynamic array
	at the position specified by the input index.

	Returns: 0 on success or -1 if the input index is out of range
----------------------------------------------*/
int DA_SetChar(
	DArray *theArray,	/* the array to set the char at */
	int nIdx,			/* the index of the array to set the char */
	char cChar			/* the character to set */
	)
{
	if ((nIdx < 0) || (nIdx > theArray->nCapacity))
	{
		return -1;
	}
	else
	{
		theArray->szArray[nIdx] = cChar;
		return 0;
	}

}


/*----- M E S S A G E   I N F O    F U N C T I O N S ------*/

/*----------------------------------------------
	Initialize the contents of a Message Info Structure.
	Since MessageInfo contains DArrays, they must be initialized
	before use

	Returns: nothing
----------------------------------------------*/
void MI_Initialize(MessageInfo *pMsgInfo)
{
	pMsgInfo->eMsgNum = E_NoCommand;
	DA_Initialize(&pMsgInfo->daMsgText);
	DA_Initialize(&pMsgInfo->daHeaders);
	pMsgInfo->szMsgBody = "";
	pMsgInfo->nBodySize = 0;
	pMsgInfo->bUseCompression = 0;

}

/*----------------------------------------------
	Reset the contents of a Message Info Structure.

	Returns: nothing
----------------------------------------------*/
void MI_Reset(MessageInfo *pMsgInfo)
{
	pMsgInfo->eMsgNum = E_NoCommand;
	DA_Reset(&pMsgInfo->daMsgText);
	DA_Reset(&pMsgInfo->daHeaders);
	pMsgInfo->szMsgBody = "";
	pMsgInfo->nBodySize = 0;
	pMsgInfo->bUseCompression = 0;

}



/*----- C O M M O N    F U N C T I O N S ------*/


/*----------------------------------------------
	Append a header name and value to a header dynamic array

	Returns: nothing
----------------------------------------------*/
void AddHeader(
	DArray *daOutHeaders, /* the header array to append the new header to */
	char *szInHeaderName, /* the new header name */
	char *szInHeaderValue /* the new header value */
	)
{
	DA_Append(daOutHeaders, szInHeaderName, -1);
	DA_Append(daOutHeaders, ":", -1);
	DA_Append(daOutHeaders, szInHeaderValue, -1);
	DA_Append(daOutHeaders, "\r\n", -1);
}


/*----------------------------------------------
	Get the value of a header field in a header array.

  	Returns: The value of the header if found, or NULL if
			not found.
----------------------------------------------*/
char *GetHeaderValue(
	DArray *daHeaders,	/* the header array to pull the value from */
	char *szHeaderName	/* the name of the header */
	)
{
	char *szHeaderValue=NULL;	/* the return value of the header */
	int i = 0;					/* Index */
	int nBegin;					/* the start idx of the found header value */
	int nEnd;					/* the end idx of the found header value */
	int bWrongHeader = 0;

	while (i < daHeaders->nCurLen)
	{
		if (BeginsWithIgnoreCase(&daHeaders->szArray[i], szHeaderName))
		{
			/* if string begins with the header name, determine the length
			of the value */
			
			/* go to beginning of value */
			nBegin = i + strlen(szHeaderName);
			
			bWrongHeader = 0;
			while (daHeaders->szArray[nBegin] != ':')
			{
				if (daHeaders->szArray[nBegin] != ' ')
					bWrongHeader = 1;
				nBegin++;
			}
			if (bWrongHeader == 0)
			{
				nBegin++;
				nEnd = nBegin;
				while (daHeaders->szArray[nEnd] != '\r' &&
						daHeaders->szArray[nEnd] != '\n')
				{
					nEnd++;
				}
				szHeaderValue = (char *)malloc (nEnd - nBegin + 1);
				memcpy(szHeaderValue, &daHeaders->szArray[nBegin], nEnd - nBegin);
				szHeaderValue[nEnd-nBegin] = '\0';
				break;
			}
		}
		/* skip to next header */
		while (++i < daHeaders->nCurLen && daHeaders->szArray[i-1] != '\n')
		{
				;	/* Do Nothing */
		}


	}

	return szHeaderValue;

}

/*----------------------------------------------
	Get the value of a header field in a header array
	for a header of type int.

  	Returns: The value of the header if found, or -1 if
			not found.
----------------------------------------------*/
int GetIntHeaderValue(
	DArray *daHeaders,	/* the header array to pull the value from */
	char *szHeaderName	/* the name of the header */
	)
{
	char *temp;

	temp = GetHeaderValue(daHeaders, szHeaderName);
	if (temp != NULL)
	{
		return atoi(temp);
	} else {
		return -1;
	}
}


/*----------------------------------------------
	Get the value of the content length header in the
	input header array.

	Returns: The content length, or 0 if not found.
----------------------------------------------*/
int GetContentLength(DArray *daHeaders)
{
	int nContentLen=0;
	char *szContentLenText;

	szContentLenText = GetHeaderValue(daHeaders, H_CONTENT_LEN);

	if (szContentLenText != NULL)
	{
		sscanf(szContentLenText, "%d", &nContentLen);
	}
	return nContentLen;

}

/*----------------------------------------------
	Compare the begining of one string with 
	another string,	ignoring case.

	Returns: 1 if s1 begins with s2, 0 otherwise
----------------------------------------------*/
int BeginsWithIgnoreCase(
	char *s1,	/* string to check */
	char *s2)	/* string containing prefix */
{
	unsigned int i=0;

	if (!s1 || !s2 || (strlen(s1) == 0) ||
		(strlen(s2) == 0) || 
		(strlen(s1) < strlen(s2)))
	{
		return 0;
	}

	for (i=0; i < strlen(s2); i++)
	{
		if (toupper(s1[i]) != toupper(s2[i]))
		{
			return 0;
		}
	}
	return 1;
}

/*----------------------------------------------
	Compare the end of one string with 
	another string, ignoring case

	Returns: 1 if s1 ends with s2, 0 otherwise
----------------------------------------------*/
int EndsWithIgnoreCase(	
	char *s1,	/* string to check */
	char *s2)	/* string containing postfix */

{
	unsigned int i=0;
	unsigned int start=0;

	if (!s1 || !s2 || (strlen(s1) == 0) ||
		(strlen(s2) == 0) || 
		strlen(s1) < strlen(s2))
	{
		return 0;
	}

	start = strlen(s1)-strlen(s2);
	for (i=0; i < strlen(s2); i++)
	{
		if (toupper(s1[start+i]) != toupper(s2[i]))
		{
			return 0;
		}
	}
	return 1;
}


/*----------------------------------------------
	Convert the input command string to its MsgNum
	equivalent.

	Returns: The enum MsgNum that corresponds to the string, or
				E_NoCommand if it cannot find a match.
----------------------------------------------*/
enum MsgNum ConvertMsgStr2MsgNum(
	char *szCmd		/* the command as a string */
	)
{
	enum MsgNum theECmd = E_NoCommand;

	if (BeginsWithIgnoreCase(MSG_GET_DATABASES, szCmd))
	{
		theECmd = E_GetDatabases;
	} 
	else if (BeginsWithIgnoreCase(MSG_GET_FILE, szCmd))
	{
		theECmd = E_GetFile;
	}
	else if (BeginsWithIgnoreCase(MSG_GET_FILE_LIST, szCmd))
	{
		theECmd = E_GetFileList;
	}
	else if (BeginsWithIgnoreCase(MSG_CREATE_PROJECT, szCmd))
	{
		theECmd = E_CreateProject;
	}
	else if (BeginsWithIgnoreCase(MSG_ADD_FILE, szCmd))
	{
		theECmd = E_AddFile;
	}
	else if (BeginsWithIgnoreCase(MSG_CHECK_OUT_FILE, szCmd))
	{
		theECmd = E_CheckOutFile;
	}
	else if (BeginsWithIgnoreCase(MSG_CHECK_IN_FILE, szCmd))
	{
		theECmd = E_CheckInFile;
	}
	else if (BeginsWithIgnoreCase(MSG_UNDO_CHECKOUT, szCmd))
	{
		theECmd = E_UndoCheckOut;
	}
	else if (BeginsWithIgnoreCase(MSG_GET_FILE_HISTORY, szCmd))
	{
		theECmd = E_GetFileHistory;
	}
	else if (BeginsWithIgnoreCase(MSG_ADD_LABEL, szCmd))
	{
		theECmd = E_AddLabel;
	}

	else if (BeginsWithIgnoreCase(MSG_GET_PROJECT, szCmd))
	{
		theECmd = E_GetProject;
	}
	else if (BeginsWithIgnoreCase(MSG_GET_PROJECT_TREE, szCmd))
	{
		theECmd = E_GetProjectTree;
	}
	else if (BeginsWithIgnoreCase(MSG_DELETE, szCmd))
	{
		theECmd = E_Delete;
	}
	else if (BeginsWithIgnoreCase(MSG_RENAME, szCmd))
	{
		theECmd = E_Rename;
	}
	else if (BeginsWithIgnoreCase(MSG_GET_FILE_INFO, szCmd))
	{
		theECmd = E_GetFileInfo;
	}
	else if (BeginsWithIgnoreCase(MSG_SHAREFILE, szCmd))
	{
		theECmd = E_ShareFile;
	}
	else if (BeginsWithIgnoreCase(MSG_BRANCHFILE, szCmd))
	{
		theECmd = E_BranchFile;
	}
	else if (BeginsWithIgnoreCase(MSG_ADDFOLDER, szCmd))
	{
		theECmd = E_AddFolder;
	}
	else if (BeginsWithIgnoreCase(MSG_CHECKINPROJECT, szCmd))
	{
		theECmd = E_CheckInProject;
	}
	else if (BeginsWithIgnoreCase(MSG_CHECKOUTPROJECT, szCmd))
	{
		theECmd = E_CheckOutProject;
	}
	else if (BeginsWithIgnoreCase(MSG_UNDOPROJECTCHECKOUT, szCmd))
	{
		theECmd = E_UndoProjectCheckOut;
	}
	else if (BeginsWithIgnoreCase(MSG_BRANCHPROJECT, szCmd))
	{
		theECmd = E_BranchProject;
	}
	else if (BeginsWithIgnoreCase(MSG_SHAREPROJECT, szCmd))
	{
		theECmd = E_ShareProject;
	}

	return theECmd;
}


/*----------------------------------------------
	Convert the input MsgNum to its message string
	equivalent.

	Returns: The enum MsgNum that corresponds to the string, or
				E_NoCommand if it cannot find a match.
----------------------------------------------*/
char *ConvertMsgNum2MsgStr(
	enum MsgNum eMsgNum		/* the Msg Number */
	)
{
	char *szCmdStr;

	switch (eMsgNum)
	{
		case E_GetDatabases:
			szCmdStr = MSG_GET_DATABASES;
			break;
		case E_Login:
			szCmdStr = MSG_LOGIN;
			break;
		case E_GetProjectTree:
			szCmdStr = MSG_GET_PROJECT_TREE;
			break;
		case E_GetFileList:
			szCmdStr = MSG_GET_FILE_LIST;
			break;
		case E_CreateProject:
			szCmdStr = MSG_CREATE_PROJECT;
			break;
		case E_AddFile:
			szCmdStr = MSG_ADD_FILE;
			break;
		case E_GetFile:
			szCmdStr = MSG_GET_FILE;
			break;
		case E_CheckOutFile:
			szCmdStr = MSG_CHECK_OUT_FILE;
			break;
		case E_CheckInFile:
			szCmdStr = MSG_CHECK_IN_FILE;
			break;
		case E_UndoCheckOut:
			szCmdStr = MSG_UNDO_CHECKOUT;
			break;
		case E_GetFileHistory:
			szCmdStr = MSG_GET_FILE_HISTORY;
			break;
		case E_AddLabel:
			szCmdStr = MSG_ADD_LABEL;
			break;
		case E_Challenge:
			szCmdStr = MSG_CHALLENGE;
			break;
		case E_SecureChallenge:
			szCmdStr = MSG_SECURE_CHALLENGE;
			break;
		case E_ProjectTree:
			szCmdStr = MSG_PROJECT_TREE;
			break;
		case E_FileList:
			szCmdStr = MSG_FILE_LIST;
			break;
		case E_File:
			szCmdStr = MSG_FILE;
			break;
		case E_FileHistory:
			szCmdStr = MSG_FILE_HISTORY;
			break;
		case E_GetProject:
			szCmdStr = MSG_GET_PROJECT;
			break;
		case E_Delete:
			szCmdStr = MSG_DELETE;
			break;
		case E_Rename:
			szCmdStr = MSG_RENAME;
			break;
		case E_GetFileInfo:
			szCmdStr = MSG_GET_FILE_INFO;
			break;
		case E_ShareFile:
			szCmdStr = MSG_SHAREFILE;
			break;
		case E_BranchFile:
			szCmdStr = MSG_BRANCHFILE;
			break;
		case E_AddFolder:
			szCmdStr = MSG_ADDFOLDER;
			break;
		case E_CheckInProject:
			szCmdStr = MSG_CHECKINPROJECT;
			break;
		case E_UndoProjectCheckOut:
			szCmdStr = MSG_UNDOPROJECTCHECKOUT;
			break;
		case E_BranchProject:
			szCmdStr = MSG_BRANCHPROJECT;
			break;
		case E_CheckOutProject:
			szCmdStr = MSG_CHECKOUTPROJECT;
			break;
		case E_ShareProject:
			szCmdStr = MSG_SHAREPROJECT;
			break;

		case E_Success:
			szCmdStr = MSG_SUCCESS;
			break;
		case E_CheckinComplete:
			szCmdStr = MSG_CHECKIN_COMPLETE;
			break;
		case E_FileUnlocked:
			szCmdStr = MSG_FILE_UNLOCKED;
			break;
		case E_FileAdded:
			szCmdStr = MSG_FILE_ADDED;
			break;
		case E_FileLocked:
			szCmdStr = MSG_FILE_LOCKED;
			break;
		case E_FileInfo:
			szCmdStr = MSG_FILE_INFO;
			break;
		case E_NoSuchFileOrProject:
			szCmdStr = MSG_NO_SUCH_FILE_OR_PROJECT;
			break;
		case E_FileOrProjectNotLocked:
			szCmdStr = MSG_FILE_OR_PROJECT_NOT_LOCKED;
			break;
		case E_AuthFailed:
			szCmdStr = MSG_AUTH_FAILED;
			break;
		case E_MissingHeader:
			szCmdStr = MSG_MISSING_HEADER;
			break;
		case E_InvalidLabel:
			szCmdStr = MSG_INVALID_LABEL;
			break;
		case E_InvalidVersion:
			szCmdStr = MSG_INVALID_VERSION;
			break;
		case E_UnrecognizedMsg:
			szCmdStr = MSG_UNRECOGNIZED_MSG;
			break;
		case E_InvalidProtocolVersion:
			szCmdStr = MSG_INVALID_PROTOCOL_VERSION;
			break;
		case E_FileOrProjectAlreadyExists:
			szCmdStr = MSG_FILE_OR_PROJECT_ALREADY_EXISTS;
			break;
		case E_FileOrProjectAlreadyLocked:
			szCmdStr = MSG_FILE_OR_PROJECT_ALREADY_LOCKED;
			break;
		case E_UnathorizedUser:
			szCmdStr = MSG_UNATHORIZED_USER;
			break;
		case E_ServerExpired:
			szCmdStr = MSG_SERVER_EXPIRED;
			break;
		case E_OperationFailed:
			szCmdStr = MSG_OPERATION_FAILED;
			break;
		case E_NoCommand:
			szCmdStr = MSG_NO_COMMAND;
			break;
		case E_KeyNotFound:
			szCmdStr = MSG_KEY_NOT_FOUND;
			break;
		case E_KeyExpired:
			szCmdStr = MSG_KEY_EXPIRED;
			break;
		case E_InvalidKey:
			szCmdStr = MSG_INVALID_KEY;
			break;
		case E_BadIniFile:
			szCmdStr = MSG_BAD_INI_FILE;
			break;
		case E_MissingSSAuto:
			szCmdStr = MSG_MISSING_AUTO;
			break;
		case E_MissingCrypto:
			szCmdStr = MSG_MISSING_CRYPTO;
			break;
		case E_CheckinMergeFailed:
			szCmdStr = MSG_CHECKIN_MERGE_FAILED;
			break;
		case E_FailedDatabaseOpen:
			szCmdStr = MSG_FAILED_DATABASE_OPEN;
			break;
		default:
			szCmdStr = "Unknown Server Number";
			break;

	}

	return szCmdStr;
}

/*----------------------------------------------
	Extracts the file name part of a file path

	Returns: The pointer to the beginning of the file
		path.  This function does NOT create a new
		string, but returns a pointer within the existing
		string.
----------------------------------------------*/
char *ExtractFileName(char szFilePath[])
{
	int idx;

	if (szFilePath == NULL)
		return NULL;

	idx = strlen(szFilePath);
	while (idx >= 0)
	{
		if (szFilePath[idx] == '/' || szFilePath[idx] == '\\')
		{
			idx++;
			break;
		}
		idx--;
	}
	if (idx < 0)
		idx++;
	return szFilePath + idx;
	
}


/*----------------------------------------------
	Set a file's permissions to ReadOnly

	Returns 0 on success, errno on error
----------------------------------------------*/
int MakeFileReadOnly(char *szFilePath)
{
	int s;
	STAT sb;
	memset(&sb, 0, sizeof(STAT));

	if (stat(szFilePath, &sb))
		return -1;

#ifdef WIN32
	if (sb.st_mode & S_IWRITE)
		sb.st_mode ^= S_IWRITE;
#else
	if (sb.st_mode & S_IWUSR)
		sb.st_mode ^= S_IWUSR;
	if (sb.st_mode & S_IWGRP)
		sb.st_mode ^= S_IWGRP;
	if (sb.st_mode & S_IWOTH)
		sb.st_mode ^= S_IWOTH;
#endif
	
	s = chmod(szFilePath, sb.st_mode);
	if (s != 0)
	{
		printf("Warning:  Could not make file %s read-only\n", szFilePath);
	}

	return s;

/*
	int s;

	s = chmod(szFilePath, S_IREAD);
	if (s != 0)
	{
		printf("Warning:  Could not make file %s read-only\n", szFilePath);
	}

	return s;
*/
}

/*----------------------------------------------
	Set a file's permissions to ReadWrite

	Returns 0 on success, errno on error
----------------------------------------------*/
int MakeFileReadWrite(char *szFilePath)
{
	int s = 0;
	STAT sb;
	memset(&sb, 0, sizeof(STAT));

	if (stat(szFilePath, &sb))
		return -1;

	if (!(sb.st_mode & S_IWRITE))
	{
		sb.st_mode |= S_IWRITE;
	
		s = chmod(szFilePath, sb.st_mode);
		if (s != 0)
		{
			printf("Warning:  Could not make file %s read-write\n", szFilePath);
		}
	}

	return s;
}


/*----------------------------------------------
	Copy the input file to the current working directory,
	then delete the input file from its old location.

	Return 0 on success, errno on error
----------------------------------------------*/
int MoveFileToCurDir(char *szFromFile, char *szToFile)
{
	/*** NOTE: Do this in platform independent way ****/

	FILE *pFromFile;		/* The source file pointer */
	FILE *pToFile;			/* The destination file pointer */
	char fileBuf[1024];		/* buffer for file content copying */
	int nBytesRead;			/* the number of bytes read from the file */



	/* Open the new file for writing */
	pToFile=fopen(szToFile, "wb");

	/* if open failed, attempt to make the file writable, and 
		attempt to open file again */
	if (pToFile == NULL)
	{
		MakeFileReadWrite(szToFile);
		pToFile=fopen(szToFile, "wb");
	}

	/* if open failed, output error message and exit */
	if (pToFile == NULL)
	{
		printf("Could not open file %s.  Error Num=%d\n", szToFile, errno);
		return -1;
	}

	/* open the source file */
	pFromFile=fopen(szFromFile, "rb");
	/* if open failed, output error message and exit */
	if (pFromFile == NULL)
	{
		printf("Could not open file %s.  Error Num=%d\n", szFromFile, errno);
		return -1;
	}


	/* Copy the contents of the file */
	
	do 
	{
		nBytesRead = fread(fileBuf, 1, 1024, pFromFile);
		if (nBytesRead > 0)
		{
			fwrite(fileBuf, 1, nBytesRead, pToFile);
		}
	} while (nBytesRead > 0);


	/* Close the files */
	if (fclose(pFromFile) != 0)
	{
		printf("Could not close file %s.  Error Num=%d\n", szFromFile, errno);
		return -1;
	}

	if (fclose(pToFile) != 0)
	{
		printf("Could not close file %s.  Error Num=%d\n", szToFile, errno);
		return -1;
	}

	/* delete the old file */
	unlink(szFromFile);


	return 0;


}

/*----------------------------------------------
	Change to the input directory, creating intermediate
	directories as necessary.

	Return 0 on success, errno on error
----------------------------------------------*/
int ChangeDir(char *szInDirPath)
{
	int sidx;			/* start index for a subdir */
	int eidx;			/* end index for a subdir */
	int maxidx;			/* length of dir path */
	char subdir[512];

	if (szInDirPath == NULL)
		return 0;

	maxidx = strlen(szInDirPath);
	sidx = eidx = 0;
	/* 
		find each subdir in path, and cd to each indvidual
		directory, creating it if it doesn't exist 
	*/
	while (eidx < maxidx)
	{
		/* If found a subdir */
		if (eidx == maxidx-1 ||
			szInDirPath[eidx] == '/' ||
			szInDirPath[eidx] == '\\')
		{
			/* Check if directory exists, and create if it doesn't */
			strncpy(subdir, szInDirPath+sidx, eidx - sidx + 1);
			subdir[eidx - sidx + 1] = '\0';
			if (access(subdir, 00) != 0 && errno == ENOENT)
			{
#ifdef WIN32
				if (mkdir(subdir) != 0)
#else
				if (mkdir(subdir, 0755) != 0)
#endif /* WIN32 */
				{
					printf("Could not create directory %s.  Error Num=%d\n", subdir, errno);
					return errno;
				}
			}

			/* Change to the new directory */
			if (chdir(subdir) != 0)
			{
				printf("Could not change to working directory %s.  Error Num=%d\n", subdir, errno);
				return errno;
			}

			/* reset vars to look for next directory */
			sidx = eidx + 1;
		}
		eidx++;
	}


	return 0;

}


char* ConvertServerName(const char* pServer)
{
	int i = 0;
	int numFound = 0;
	int pos = 0;
	char* pNewServer;
	int tmp = 0;

	if (!pServer)
		return NULL;

	pNewServer = (char *)malloc(13);

	pNewServer[0] = 0;

	while (pServer[i])
	{
		if (pServer[i] >= '0' && pServer[i] <= '9')
		{
			++numFound;
		}
		else if (pServer[i] == '.')
		{
			/* bogus ip numbers ... */
			if (numFound > 3)
			{
				free(pNewServer);
				return (char*)pServer;
			}

			tmp = numFound;
			while (tmp < 3)
			{
				pNewServer[pos++] = '0';
				++tmp;
			}

			tmp = numFound;

			while (tmp > 0)
			{
				pNewServer[pos++] = pServer[i - tmp];
				tmp--;
			}

			numFound = 0;
		}
		else
		{
			free(pNewServer);
			return (char *)pServer;
		}

		++i;
	}

	if (pos < 12)
	{
		/* bogus ip numbers ... */
		if (numFound > 3)
		{
			free(pNewServer);
			return (char*)pServer;
		}

		tmp = numFound;
		while (tmp < 3)
		{
			pNewServer[pos++] = '0';
			++tmp;
		}

		tmp = numFound;

		while (tmp > 0)
		{
			pNewServer[pos++] = pServer[i - tmp];
			tmp--;
		}
	}

	pNewServer[12] = 0;

	return pNewServer;
}



#ifdef WIN32

char *GetFirstFile(long *pHandle, char *pFileSpec)
{
	struct _finddata_t pFileInfo;
	char szFileSpec[2048];
	char* p = NULL;

	sprintf(szFileSpec, "%s%c%s", pFileSpec, SLASH, "*");

    if((*pHandle = _findfirst(szFileSpec, &pFileInfo)) == -1L)
	{
		return NULL;
	}

	p = (char*) malloc(strlen(pFileInfo.name) + 1);
	strcpy(p, pFileInfo.name);
	return p;
}


char *GetNextFile(long *pHandle)
{
	struct _finddata_t pFileInfo;
	char* p = NULL;

	if(_findnext(*pHandle, &pFileInfo) != 0)
	{
		return NULL;
	}

	p = (char*) malloc(strlen(pFileInfo.name) + 1);
	strcpy(p, pFileInfo.name);
	return p;
}

#else

char *GetFirstFile(void **pHandle, char *pFileSpec)
{
	struct dirent *pDirEnt = NULL;
	char* p = NULL;

	*pHandle = (void*)opendir(pFileSpec);

    if((pDirEnt = readdir((DIR *)(*pHandle))) == NULL)
	{
		return NULL;
	}

	p = (char*) malloc(strlen(pDirEnt->d_name) + 1);
	strcpy(p, pDirEnt->d_name);
	return p;
}


char *GetNextFile(void **pHandle)
{
	struct dirent *pDirEnt = NULL;
	char* p = NULL;

    if((pDirEnt = readdir((DIR *)(*pHandle))) == NULL)
	{
		return NULL;
	}

	p = (char*) malloc(strlen(pDirEnt->d_name) + 1);
	strcpy(p, pDirEnt->d_name);
	return p;
}

#endif
