/*******************************************************

	FILE: CLnet.c

	DESC: This file contains definitions for Network 
		related functions.
 
        Copyright 1998-2002 SourceGear Corporation.
        All Rights Reserved. 

*******************************************************/

#include <stdio.h>
#include <time.h>

#ifndef BUILD_STANDARD
#include "CLcrypto.h"
#endif  /* !BUILD_STANDARD */

#include "zlib.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>

#include "CLcommon.h"
#include "CLnet.h"
#include "CLTree.h"


#ifdef WIN32
#include <direct.h>
#include <winsock.h>
#endif /* WIN32 */

#ifdef UNIX
#include <unistd.h>
#include <dirent.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

/* Solaris doesn't define INADDR_NONE */
#ifndef INADDR_NONE
#define INADDR_NONE (unsigned long) -1
#endif

#endif /* UNIX */

extern int errno;

#define NET_BUF_SIZE 512

/* Function prototypes local to this file */
void Net_Error(char *ErrMsg);
int Net_Read(int socket, char *buf, int len);
int Net_Write(int socket, char *buf, int len);
int Net_ReadHeaders(int nInSocket, DArray *daInArray, RequestInfo *pReqInfo);

/*------------------------------------------------------
	Open a socket and connect to the input server and 
	the input port.  

	Returns: The open socket on success or a negative 
	number if an error occured, which was written to stdout.
-------------------------------------------------------*/
int Net_ConnectToServer(
	char *szServer,		/* server to connect to */
	int nPort			/* port on server to connect to */
	)
{
	int nSocket;				/* The socket to open */
	struct sockaddr_in sin;		/* an Internet endpoint address */
	struct hostent *phe;		/* pointer to host information entry */

#ifdef WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	wVersionRequested = MAKEWORD( 2, 0 );
 
	if (WSAStartup( wVersionRequested, &wsaData ) != 0 ) 
	{
		/* Tell the user that we couldn't find a usable */
		/* WinSock DLL. */
		printf("Couldn't find a usable WinSock DLL\n");
		return -1 ;
	}
#endif

	/* Reserve a socket */
	nSocket = socket(PF_INET, SOCK_STREAM, 0);
	if (nSocket < 0)
	{
		Net_Error("Can't create socket.");
		return -1 ;
	}

	/* Initialize sin structure */
	memset( &sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;

	/* Convert port number to correct network byte order */
	sin.sin_port = htons(( u_short ) nPort );
	if( sin.sin_port == 0 )
	{
		Net_Error( "Couldn't convert port to network byte order." );
		return -1 ;
	}

	/* Convert IP Address */
	if ( (sin.sin_addr.s_addr = inet_addr(szServer)) == INADDR_NONE)
	{
		/* Map host name to IP address, allowing for dotted decimal */
		if ((phe = gethostbyname(szServer)))
		{
			memcpy(&sin.sin_addr, phe->h_addr, phe->h_length);
		}
		else
		{
			Net_Error("Can't get server host entry");
			return -1 ;
		}
	}

	/* connect the socket to the input server */
	if (connect(nSocket, (struct sockaddr *)&sin, sizeof(sin)) < 0)
	{
		Net_Error("Can't connect to server.");
		return -1 ;
	}

	/* Return sucess code */
	return nSocket;
}

/*------------------------------------------------------
	Close the server connection  

	Returns: Nothing
-------------------------------------------------------*/
void Net_CloseServerConnection (
		int nSocket,		/* an open socket */
		RequestInfo *pReqInfo
		)
{
	/* First, send the exit command */
	MessageInfo	exitMsg;
	MI_Initialize(&exitMsg);
	exitMsg.eMsgNum = E_Exit;
	Net_SendRequest(nSocket, &exitMsg, pReqInfo);


	if (shutdown(nSocket, 2) != 0)
	{
		Net_Error("Can't shutdown socket.");
	}

#ifdef WIN32
	closesocket(nSocket);
	WSACleanup ();
#endif
}


/*------------------------------------------------------
	Read one line of text from the input socket and append 
	it to the OutArray parameter. 

	*IMPORTANT* Don't use this function for reading message 
	bodies, because a message body is not guaranteed to end 
	with a "\r\n" (and this function is really slow). 
	To read the body, use the content-length header to get 
	the number of bytes in the message body.

	Returns: The number of bytes read, or a negative
		number on error.
-------------------------------------------------------*/
int NET_ReadNonBodyLine(
	DArray *daInput,	/* input array to parse */
	DArray *daOutArray	/* output array to append line to */
	)
{
	char *pBuf;
	int	 nRespLen=0;	/* the total length of the line */

	pBuf = strstr(daInput->szArray, "\r\n");
	if (pBuf == NULL)
	{
		Net_Error("Error during attempt to read message line. (RNBL1)");
		return -1 ;
	} else {
		pBuf += 2;
		nRespLen = pBuf - daInput->szArray;
		DA_Append(daOutArray, daInput->szArray, nRespLen);
		DA_Assign(daInput, pBuf, daInput->nCurLen - nRespLen); 
	}

	/* return the length of the array */
	return  nRespLen;
}


/*------------------------------------------------------
	Read the response code from the input socket, and 
	place it in the output parameter.

	Returns: 0 on success, negative number on error.
-------------------------------------------------------*/
int Net_GetRespCode(
	DArray *daInput,
	enum MsgNum *nOutRespCode,	/* response code placed here  - 0 if error */
	DArray *daOutRespBuf /* optional response text placed here */
	)
{
	int	 nRespLen = 0;	/* the total length of the response line */

	DA_Reset(daOutRespBuf);

	/* Read the first line from the socket.
	*/
	nRespLen=NET_ReadNonBodyLine(daInput, daOutRespBuf);
	if (nRespLen <= 0)
	{
		DA_Assign(daOutRespBuf, "No Response Code Read", -1);
		*nOutRespCode = E_NoCommand;
		return -1;
	}

	/* strip off the carriage return/line feed */
	DA_SetChar(daOutRespBuf,nRespLen-2,'\0');

	/* scan for the output response code - it is the first
		integer on the line
	*/
	if (sscanf(daOutRespBuf->szArray, "%d", nOutRespCode) < 1)
	{
		*nOutRespCode = E_NoCommand;
	}

	return 0;

}

/*------------------------------------------------------
	Read the headers from the input socket, and 
	place them in the output parameter.

	Returns: 0 on success, negative number on error.
-------------------------------------------------------*/
int Net_ParseHeaders( 
	DArray *daInput,		/* Input array to parse */
	DArray *daOutHeaders	/* headers placed here */
	)
{
	int	 nLen=0;				/* the length of each header line */
	int	 nTotalLen=0;			/* the total length of all headers */


	DA_Reset(daOutHeaders);

	/* Read lines from the socket until we hit an empty line.  For
		each line, create a new header and add to the output array.
	*/
	while ((nLen=NET_ReadNonBodyLine(daInput, daOutHeaders)) > 2)
	{
		nTotalLen += nLen;
	}

	return 0;
}

/*------------------------------------------------------
	Read the message body from the input socket, and 
	place it in the output parameter.

	Returns: the length of the body on success, negative 
		number on error.
-------------------------------------------------------*/
int Net_GetMsgBody(
	int nInSocket, 			/* an open socket to read from */
	MessageInfo *pMsgInfo,
	RequestInfo *pReqInfo
	)
{
	int	 nLen=0;		/* the actual # of bytes read per try */
	int	 nTotalBytesRead=0;	/* the total # of bytes read */
	int i = 0;
	int nContentLen = 0;
	int nOrigLen = 0;
	char *szOutText;

#ifndef BUILD_STANDARD
	char IV[CRYPTO_HEADER_IV_SIZE + 1];
	char szLengthEnc[CRYPTO_HEADER_LENGTH_SIZE + 1];
	char szLength[CRYPTO_HEADER_LENGTH_SIZE + 1];
	int nSizeToDecrypt = 0;
	int remainder;
	char *pEncBlock;
	int nRespLen = 0;
#endif /* !BUILD_STANDARD */

	nContentLen=GetContentLength(&pMsgInfo->daHeaders);

#ifndef BUILD_STANDARD
	if (pReqInfo->bUseEncrypt == 0)
	{
#endif /* !BUILD_STANDARD */
		szOutText = (char *)malloc(nContentLen+1);
	
		while (nTotalBytesRead < nContentLen)
		{
			nLen = Net_Read(nInSocket, szOutText+nTotalBytesRead, 
				nContentLen-nTotalBytesRead);
			if (nLen < 0)
			{
				Net_Error("Net error during attempt to read message body.");
				return -1;
			} 
			
			nTotalBytesRead += nLen;
		}
		
		if (nTotalBytesRead != nContentLen)
		{
			printf("Read different number of bytes in message body than expected.\n");
			printf("Expected Body Length: %d\tActual bytes read: %d\n", nContentLen, nTotalBytesRead);
		}
		/* ensure the body text is null terminted */
		szOutText[nTotalBytesRead] = '\0';


#ifndef BUILD_STANDARD
	} else {
		nLen = Net_Read(nInSocket, IV, CRYPTO_HEADER_IV_SIZE);
		if (nLen < CRYPTO_HEADER_IV_SIZE)
		{
			Net_Error("Net error during attempt to read message line. (GMB1)");
			return -1 ;
		} 

		nLen = Net_Read(nInSocket, szLengthEnc, CRYPTO_HEADER_LENGTH_SIZE);
		if (nLen < CRYPTO_HEADER_LENGTH_SIZE)
		{
			Net_Error("Net error during attempt to read message line. (GMB2)");
			return -1 ;
		} 

		Decrypt(szLengthEnc, szLength, CRYPTO_HEADER_LENGTH_SIZE, pReqInfo->bfKey, IV); 

		while ((i < CRYPTO_HEADER_LENGTH_SIZE) && (szLength[i] != PAD_CHAR))
			i++;
		
		szLength[i] = '\0';
		
		nRespLen = atoi(szLength);
		
		/* determine how many total bytes will be arriving including the pad? */
		
		nSizeToDecrypt = nRespLen;
		
		if (nSizeToDecrypt < CRYPTO_BLOCK_SIZE)
		{
			nSizeToDecrypt = CRYPTO_BLOCK_SIZE;
		}
		else
		{
			remainder = nSizeToDecrypt % CRYPTO_BLOCK_SIZE;
			
			if (remainder > 0)
				nSizeToDecrypt += CRYPTO_BLOCK_SIZE - remainder;
		}

		pEncBlock = (char *)malloc(nSizeToDecrypt);
		szOutText = (char *)malloc(nSizeToDecrypt + 1);


		while (nTotalBytesRead < nSizeToDecrypt)
		{
			nLen = Net_Read(nInSocket, pEncBlock+nTotalBytesRead, 
				nSizeToDecrypt-nTotalBytesRead);
			if (nLen < 0)
			{
				Net_Error("Net error during attempt to read message body.");
				return -1;
			} 
			
			nTotalBytesRead += nLen;
		}

		Decrypt(pEncBlock, szOutText, nSizeToDecrypt, pReqInfo->bfKey, IV);

		szOutText[nRespLen] = '\0';
	}
#endif /* !BUILD_STANDARD */

	if (GetHeaderValue(&pMsgInfo->daHeaders, H_USE_COMPRESSION) != NULL)
	{
		nOrigLen = GetIntHeaderValue(&pMsgInfo->daHeaders, H_ORIGINAL_SIZE);
		pMsgInfo->szMsgBody = (char *)malloc(nOrigLen + 1);

		uncompress((unsigned char *)pMsgInfo->szMsgBody, (long unsigned int *)&nOrigLen, (unsigned char *)szOutText, nContentLen);
		pMsgInfo->szMsgBody[nOrigLen] = '\0';
		pMsgInfo->nBodySize = nOrigLen;
	} else {
		pMsgInfo->szMsgBody = szOutText;
		pMsgInfo->nBodySize = nContentLen;
	}

	return nLen;
}

/*------------------------------------------------------
	Read the message body from the input socket, and 
	place it in a file.

	Returns: 0 on success, -1 on error.
-------------------------------------------------------*/
int Net_SendBodyToFile(
	int			nInSocket, 		/* an open socket to read from */
	MessageInfo	*pMsgInfo,
	char		*szFileName,	/* File name to place body in */
	char		*szWorkDir,		/* Working directory, or NULL if none */
	int			bOverride,		/* whether to overwrite file if already exists */
	RequestInfo	*pReqInfo
	)
{
	char fileBuf[NET_BUF_SIZE];	/* the buf that will hold the file */
	int nBytesToRead;		/* the number of bytes to read into the buffer */
	int nBytesRead;			/* the number of bytes actually read into the buffer */
	int nBytesLeft;			/* the number of bytes remaining to be read */
	FILE *pFile;			/* The file pointer */
	int bWorkdirError=0;		/* There was a problem with the work dir */
	int i = 0;
	int nContentLen = 0;
	int nOrigLen = 0;
	char *pCompBuf;
	DArray daFileBuf;
	char* pszNewFile = NULL;
	
#ifndef BUILD_STANDARD
	char IV[CRYPTO_HEADER_IV_SIZE + 1];
	char szLengthEnc[CRYPTO_HEADER_LENGTH_SIZE + 1];
	char szLength[CRYPTO_HEADER_LENGTH_SIZE + 1];
	int nSizeToDecrypt = 0;
	int remainder;
	char *pOutBlock;
	int nRespLen = 0;
#endif /* !BUILD_STANDARD */

	DA_Initialize(&daFileBuf);
	nContentLen = GetContentLength(&pMsgInfo->daHeaders);

	/* Change to the work dir if it exists */
	if (szWorkDir != NULL)
	{
		STAT st;

		ChangeDir(szWorkDir);
		
		if (stat(szWorkDir, &st))
		{
			/* Note: Don't return right away - we need to pull the rest
			   of the bytes off the net in case another command is received */
			printf("Could not change to working directory %s.  Error Num=%d\n", szWorkDir, errno);
			bWorkdirError = 1;
		}
	}
	
	/* Open the file for writing */
	if (bWorkdirError == 1)
	{
		pFile = NULL;
	} else {
		int len = strlen(szFileName) + 1;

		if (szWorkDir)
			len += strlen(szWorkDir) + 1;

		pszNewFile = (char*) malloc(len);

		if (!pszNewFile)
			return -1;
		
		if (szWorkDir)
			sprintf(pszNewFile, "%s%c%s", szWorkDir, SLASH, szFileName);
		else
			sprintf(pszNewFile, "%s", szFileName);
	   
		pFile=fopen(pszNewFile, "wb");

		/* if open failed, but override is set, attempt to make
			the file writable, and attempt to open file again */
		if (pFile == NULL && bOverride)
		{
			MakeFileReadWrite(pszNewFile);
			pFile=fopen(pszNewFile, "wb");
		}
	}

	if (pszNewFile)
		free(pszNewFile);

	/* if open failed, output error message and exit */
	if (pFile == NULL && bWorkdirError == 0)
	{
		/* Note: Don't return yet, because we still need to read all 
		 the bytes off the socket in case another command is sent
		 to the server. 
		*/
		printf("Could not open file %s.  Error Num=%d\n", szFileName, errno);
	}

#ifndef BUILD_STANDARD
	if (pReqInfo->bUseEncrypt == 0)
	{
#endif /* !BUILD_STANDARD */
		if (GetHeaderValue(&pMsgInfo->daHeaders, H_USE_COMPRESSION) != NULL)
		{
			nOrigLen = GetIntHeaderValue(&pMsgInfo->daHeaders, H_ORIGINAL_SIZE);
			nBytesLeft = nContentLen;
			while (nBytesLeft > 0)
			{
				if (nBytesLeft > NET_BUF_SIZE)
				{
					nBytesToRead = NET_BUF_SIZE;
				} else {
					nBytesToRead = nBytesLeft;
				}
				
				nBytesRead = Net_Read(nInSocket, fileBuf, nBytesToRead);
				if (nBytesRead < 0)
				{
					Net_Error("Net error during attempt to read message body. (SBTF3)");
					return -1;
				} 
				
				DA_Append(&daFileBuf, fileBuf, nBytesRead);
				
				nBytesLeft = nBytesLeft - nBytesRead;
			}

			pCompBuf = (char *)malloc(nOrigLen);
			uncompress((unsigned char *)pCompBuf, (unsigned long int *)&nOrigLen, (unsigned char *)daFileBuf.szArray, daFileBuf.nCurLen);

			if (pFile != NULL && nOrigLen > 0)
			{
				fwrite(pCompBuf, 1, nOrigLen, pFile);
			}

		} else {

			/* Read the file in chunks */
			nBytesLeft = nContentLen;
			while (nBytesLeft > 0)
			{
				if (nBytesLeft > NET_BUF_SIZE)
				{
					nBytesToRead = NET_BUF_SIZE;
				} else {
					nBytesToRead = nBytesLeft;
				}
				
				nBytesRead = Net_Read(nInSocket, fileBuf, nBytesToRead);
				if (nBytesRead < 0)
				{
					Net_Error("Net error during attempt to read message body.");
					return -1;
				} 
				if (pFile != NULL && nBytesRead > 0)
				{
					fwrite(fileBuf, 1, nBytesRead, pFile);
				}
				nBytesLeft = nBytesLeft - nBytesRead;
			}
		}
#ifndef BUILD_STANDARD
	} else {
		if (nContentLen > 0)
		{
			nBytesRead = Net_Read(nInSocket, IV, CRYPTO_HEADER_IV_SIZE);
			if (nBytesRead < CRYPTO_HEADER_IV_SIZE)
			{
				Net_Error("Net error during attempt to read message line. (SBTF1)");
				return -1 ;
			} 
			
			nBytesRead = Net_Read(nInSocket, szLengthEnc, CRYPTO_HEADER_LENGTH_SIZE);
			if (nBytesRead < CRYPTO_HEADER_LENGTH_SIZE)
			{
				Net_Error("Net error during attempt to read message line. (SBTF2)");
				return -1 ;
			} 
			
			Decrypt(szLengthEnc, szLength, CRYPTO_HEADER_LENGTH_SIZE, pReqInfo->bfKey, IV); 
			
			while ((i < CRYPTO_HEADER_LENGTH_SIZE) && (szLength[i] != PAD_CHAR))
				i++;
			
			szLength[i] = '\0';
			
			nRespLen = atoi(szLength);
			
			/* determine how many total bytes will be arriving including the pad? */
			
			nSizeToDecrypt = nRespLen;
			
			if (nSizeToDecrypt < CRYPTO_BLOCK_SIZE)
			{
				nSizeToDecrypt = CRYPTO_BLOCK_SIZE;
			}
			else
			{
				remainder = nSizeToDecrypt % CRYPTO_BLOCK_SIZE;
				
				if (remainder > 0)
					nSizeToDecrypt += CRYPTO_BLOCK_SIZE - remainder;
			}
			
			pOutBlock = (char *)malloc(nSizeToDecrypt);
			
			nBytesLeft = nSizeToDecrypt;
			while (nBytesLeft > 0)
			{
				if (nBytesLeft > NET_BUF_SIZE)
				{
					nBytesToRead = NET_BUF_SIZE;
				} else {
					nBytesToRead = nBytesLeft;
				}
				
				nBytesRead = Net_Read(nInSocket, fileBuf, nBytesToRead);
				if (nBytesRead < 0)
				{
					Net_Error("Net error during attempt to read message body. (SBTF4)");
					return -1;
				} 
				
				DA_Append(&daFileBuf, fileBuf, nBytesRead);
				
				nBytesLeft = nBytesLeft - nBytesRead;
			}
			
			Decrypt(daFileBuf.szArray, pOutBlock, nSizeToDecrypt, pReqInfo->bfKey, IV);

			if (GetHeaderValue(&pMsgInfo->daHeaders, H_USE_COMPRESSION) != NULL)
			{
				nOrigLen = GetIntHeaderValue(&pMsgInfo->daHeaders, H_ORIGINAL_SIZE);
				pCompBuf = (char *)malloc(nOrigLen);

				uncompress((unsigned char *)pCompBuf, (unsigned long int *)&nOrigLen, (unsigned char *)pOutBlock, nContentLen);

				free(pOutBlock);
				pOutBlock = pCompBuf;
				nRespLen = nOrigLen;
			}

			if (pFile != NULL && nRespLen > 0)
			{
				fwrite(pOutBlock, 1, nRespLen, pFile);
			}
		}

	}
#endif /* !BUILD_STANDARD */


	/* now return if the file was never opened */
	if (pFile == NULL)
	{
		return -1;
	}

	/* Close the file */
	if (fclose(pFile) != 0)
	{
		printf("Could not close file %s.  Error Num=%d\n", szFileName, errno);
		return -1;
	}

	return 0;
}



/*------------------------------------------------------
	Get the response code, headers, and body of a server
	response message, placing it in the output parameter.

	Note that this should not be used when the expected
	response contains a file of unknown length, since
	the entire response is pulled into RAM.

	Returns: 0 on success, negative number on error.
-------------------------------------------------------*/
int Net_GetEntireResponse(
	int nInSocket, 				/* an open socket to read from */
	MessageInfo *pOutMsgInfo,	/* place response here */
	RequestInfo *pReqInfo
	)
{
	return Net_GetResponse(nInSocket, pOutMsgInfo, pReqInfo, NULL, NULL, 0);
}


/*------------------------------------------------------
	Get the response code, headers, and body of a server
	response message.  Always place the response code and
	headers in the MessageInfo struct, and place the message
	body there too if no file name is specified. If a file
	name is specified, output the message body to the input
	file.

	Returns: 0 on success, negative number on error.
-------------------------------------------------------*/
int Net_GetResponse(
	int			nInSocket, 		/* an open socket to read from */
	MessageInfo *pOutMsgInfo,	/* place response code & headers here */
	RequestInfo *pReqInfo,
	char		*szFileName,	/* File name to place body in, or NULL to place */
								/* body in the MessageInfo struct */
	char		*szWorkingDir,	/* Working directory, or NULL if none */
	int			bOverride		/* whether to overwrite file if already exists */
	)
{
	int nHeaders;			/* number of headers */
	int nRetVal;			/* Generic return value */
	int nContentLength;		/* the length of the body */
	DArray daInput;

	DA_Initialize(&daInput);

	
	nRetVal = Net_ReadHeaders(nInSocket, &daInput, pReqInfo);
	if (nRetVal <= 0)
	{
		return nRetVal;
	}

	/* Get the response code */
	nRetVal = Net_GetRespCode(&daInput,&pOutMsgInfo->eMsgNum,
		&pOutMsgInfo->daMsgText);
	if (nRetVal != 0)
	{
		return nRetVal;
	}

	/* Get the headers */
	nHeaders = Net_ParseHeaders(&daInput,&pOutMsgInfo->daHeaders);
	if (nHeaders < 0)
	{
		return nRetVal;
	}
	/* Get the content length from the headers */
	nContentLength=GetContentLength(&pOutMsgInfo->daHeaders);

#if 0
	if (szFileName != NULL)
#endif
	/* KLUDGE: TODO: This function should not have to know
	which type of message it to decide whether to send the 
	body to a file or to the message structure.  However, we
	don't want an error message to create a file, yet, we do
	want 0 length file gets to create a file.  What really
	needs to be done is to have the net functions create a 
	temporary file, and pass that back to the caller via a
	new parameter in MessageInfo.  Then the higher level
	functions can decide what to do with the file
	*/
	if (szFileName != NULL && pOutMsgInfo->eMsgNum < 300)
	{
		/* Place message body into file */
		nRetVal = Net_SendBodyToFile(nInSocket, pOutMsgInfo, 
			szFileName, szWorkingDir, bOverride, pReqInfo);
	} else if (nContentLength > 0) {
		/* Place message body into output parameter */
		nRetVal = Net_GetMsgBody(nInSocket, pOutMsgInfo, pReqInfo);
	}
	
	return nRetVal;
}


/*------------------------------------------------------
	Attaches a send response code and headers to an
	outgoing send buffer.

	Returns: 0 on success, negative number on error.
-------------------------------------------------------*/
int Net_AttachRespAndHeaders(
	MessageInfo *pInMsgInfo,/* the request info */
	DArray *daOutSendBuf,	/* the send buffer to add resp and headers to */
	int		nContentLen		/* the length of the body that will be sent */
	)
{
	static int nMsgID=1;/* Message ID increments for each message */
	char szBuf[256];	/* Buffer for additional headers */

#ifdef UNIX
	/* set the Unix EOL header */
	AddHeader(&pInMsgInfo->daHeaders, H_UNIX_EOL, "yes");
#endif /* Unix */

	/* Attach message ID header */
	sprintf(szBuf, "%s:%d\r\n", H_MESSAGE_ID, nMsgID++);
	DA_Append(&pInMsgInfo->daHeaders, szBuf, strlen(szBuf));

	/* Attach the content length header */
	sprintf(szBuf, "%s:%d\r\n", H_CONTENT_LEN, nContentLen);
	DA_Append(&pInMsgInfo->daHeaders, szBuf, strlen(szBuf));

	/* Create the send message buffer */
	DA_Initialize(daOutSendBuf);

	/* Add the response code to the message buffer */
	sprintf(szBuf, "%d\r\n", pInMsgInfo->eMsgNum);
	DA_Assign(daOutSendBuf, szBuf, -1);

	/* Add the headers to the message buffer */
	DA_Append(daOutSendBuf, pInMsgInfo->daHeaders.szArray, -1);
	DA_Append(daOutSendBuf, "\r\n", -1);

	return 0;

}


/*------------------------------------------------------
	Send an entire request to a server.

	Returns: 0 on success, negative number on error.
-------------------------------------------------------*/
int Net_SendRequest(
	int nSocket,  			/* an open socket to read from */
	MessageInfo *pMsgInfo,	/* the request to send */
	RequestInfo *pReqInfo
	)
{
	DArray daBuffer;	/* The send buffer */
	int nContentLen;	/* The content length of the message */
	int nCompLen;
	int i = 0;
	char *szBody = NULL;
	char szOrigLen[16];

#ifndef BUILD_STANDARD
	DArray daEncBuf;
	DArray *daTemp;
	unsigned char IV[CRYPTO_HEADER_IV_SIZE + 1];
	char MsgLen[CRYPTO_HEADER_LENGTH_SIZE + 1];
#endif  /* !BUILD_STANDARD */


	nContentLen = pMsgInfo->nBodySize;

	if ((nContentLen > 0) && (pMsgInfo->bUseCompression == 1))
	{
		sprintf(szOrigLen, "%d", nContentLen);
		AddHeader(&pMsgInfo->daHeaders, H_ORIGINAL_SIZE, szOrigLen);
		nCompLen = (int)(nContentLen * 1.001 + 16);
		szBody = (char *)malloc(nCompLen);
		compress((unsigned char *)szBody, (unsigned long int *)&nCompLen, (unsigned char *)pMsgInfo->szMsgBody, nContentLen);
		szBody[nCompLen] = '\0';
		nContentLen = nCompLen;
	} else {
		szBody = pMsgInfo->szMsgBody;
	}

	Net_AttachRespAndHeaders(pMsgInfo, &daBuffer, nContentLen);

#ifndef BUILD_STANDARD
	if (pReqInfo->bUseEncrypt == 1)
	{
		GetRandomNumber(IV, CRYPTO_HEADER_IV_SIZE);

		IV[CRYPTO_HEADER_IV_SIZE] = 0;

		DA_Initialize(&daEncBuf);
		DA_Assign(&daEncBuf, (char *)IV, CRYPTO_HEADER_IV_SIZE);

		sprintf(MsgLen, "%d", daBuffer.nCurLen);
		for (i = strlen(MsgLen); i < CRYPTO_HEADER_LENGTH_SIZE; i++)
			MsgLen[i] = PAD_CHAR;

		MsgLen[CRYPTO_HEADER_LENGTH_SIZE] = 0;

		daTemp = Encrypt(MsgLen, CRYPTO_HEADER_LENGTH_SIZE, pReqInfo->bfKey, IV);
		DA_Append(&daEncBuf, daTemp->szArray, daTemp->nCurLen);

		DA_Reset(daTemp);
		free(daTemp);

		daTemp = Encrypt(daBuffer.szArray, daBuffer.nCurLen, pReqInfo->bfKey, IV);
		DA_Append(&daEncBuf, daTemp->szArray, daTemp->nCurLen);

		DA_Reset(daTemp);
		free(daTemp);

		DA_Assign(&daBuffer, daEncBuf.szArray, daEncBuf.nCurLen);


		if (nContentLen > 0)
		{
			/* Encrypt message body */
			GetRandomNumber(IV, CRYPTO_HEADER_IV_SIZE);
			
			IV[CRYPTO_HEADER_IV_SIZE] = 0;
			
			DA_Reset(&daEncBuf);
			DA_Assign(&daEncBuf, (char *)IV, CRYPTO_HEADER_IV_SIZE);
			
			sprintf(MsgLen, "%d", nContentLen);
			for (i = strlen(MsgLen); i < CRYPTO_HEADER_LENGTH_SIZE; i++)
				MsgLen[i] = PAD_CHAR;
			
			MsgLen[CRYPTO_HEADER_LENGTH_SIZE] = 0;
			
			daTemp = Encrypt(MsgLen, CRYPTO_HEADER_LENGTH_SIZE, pReqInfo->bfKey, IV);
			DA_Append(&daEncBuf, daTemp->szArray, daTemp->nCurLen);
			
			DA_Reset(daTemp);
			free(daTemp);
			
			daTemp = Encrypt(szBody, nContentLen, pReqInfo->bfKey, IV);
			DA_Append(&daEncBuf, daTemp->szArray, daTemp->nCurLen);
			
			DA_Reset(daTemp);
			free(daTemp);
			
			DA_Append(&daBuffer, daEncBuf.szArray, daEncBuf.nCurLen);
		}

		DA_Reset(&daEncBuf);
	} else {
#endif  /* !BUILD_STANDARD */
	
		/* Add the body to the message buffer */
		DA_Append(&daBuffer, szBody, nContentLen);

#ifndef BUILD_STANDARD
	}
#endif  /* !BUILD_STANDARD */

#if 0
printf("Send message = %s\n", daBuffer.szArray);
#endif
	return Net_Write(nSocket, daBuffer.szArray, daBuffer.nCurLen);

}

/*------------------------------------------------------
	Send a request to a server that sends a file as the body.

	Returns: 0 on success, negative number on error.
-------------------------------------------------------*/
int Net_SendRequestAppendFile(
	int nSocket,  			/* an open socket to read from */
	MessageInfo *pMsgInfo,	/* the request to send */
	char *szFileName,		/* the name of the file to open (can be full path) */
	char *szWorkDir,			/* the working directory or NULL */
	RequestInfo *pReqInfo
	)
{
	DArray daBuffer;	/* The send buffer */
	char fileBuf[NET_BUF_SIZE];	/* the buf that will hold the file */
	int numBytes;		/* the number of bytes in the buffer */
	int nRetVal = 0;		/* Return Value */
	FILE *pFile;		/* The file pointer */
	struct stat statBuf;/* Struct that receives file length */
	int i = 0;
	char *pFileBuf = NULL;
	int nLength = 0;
	char *pCompBuf;
	int nCompLen = 0;
	char szOrigLen[16];
	char* pszNewFile = NULL;
	int len = 0;
	
#ifndef BUILD_STANDARD
	unsigned char IV[CRYPTO_HEADER_IV_SIZE + 1];
	char MsgLen[CRYPTO_HEADER_LENGTH_SIZE + 1];
	int nSizeToEncrypt = 0;
	int remainder;
	DArray daEncBuf;
	DArray *daTemp;
	char *pEncBuf;
#endif  /* !BUILD_STANDARD */


	/* Change to the work dir if it exists */
	if (szWorkDir != NULL)
	{
		STAT st;

		if (stat(szWorkDir, &st))
		{
			printf("Could not change to working directory %s.  Error Num=%d\n", szWorkDir, errno);
			return -1;
		}
	}


	if (szWorkDir)
		len += strlen(szWorkDir) + 2;
	else
		len = 1;

	len += strlen(szFileName);

	pszNewFile = (char*) malloc(len);

	if (!pszNewFile)
		return -1;
	
#ifdef WIN32
	if ((szFileName[1] == ':') || (szFileName[0] == SLASH))
#else
	if (szFileName[0] == SLASH)
#endif
	{
		sprintf(pszNewFile, "%s", szFileName);
	}
	else
	{
		if (szWorkDir)
			sprintf(pszNewFile, "%s%c%s", szWorkDir, SLASH, szFileName);
		else
			sprintf(pszNewFile, "%s", szFileName);
	}
	
	/* Open the file for reading */
	pFile=fopen(pszNewFile, "rb");
	if (pFile == NULL)
	{
		printf("Could not open file %s.  Error Num=%d\n", szFileName, errno);
		free(pszNewFile);
		return -1;
	}

	free(pszNewFile);
	
	/* Get the size of the file */
   if(fstat( fileno(pFile), &statBuf ) != 0)
   {
		printf("Could not get file size for file %s.  Error Num=%d\n", szFileName, errno);
		return -1;
   }

	nLength = statBuf.st_size;

#ifndef BUILD_STANDARD
   if ((pReqInfo->bUseEncrypt != 0) || (pMsgInfo->bUseCompression != 0))
#else
	if (pMsgInfo->bUseCompression != 0)
#endif  /* !BUILD_STANDARD */
	{

		pFileBuf = (char *)malloc(nLength);

		if (!pFileBuf)
			return -1;
		
		numBytes = fread(pFileBuf, 1, nLength, pFile);
		if (numBytes < nLength)
		{
			printf("Error reading file %s\n", szFileName);
			return -1;
		}

		if (pMsgInfo->bUseCompression != 0)
		{
			sprintf(szOrigLen, "%d", nLength);
			AddHeader(&pMsgInfo->daHeaders, H_ORIGINAL_SIZE, szOrigLen);

			nCompLen = (int)(nLength * 1.001 + 16);
			pCompBuf = (char *)malloc(nCompLen);

			compress((unsigned char *)pCompBuf, (unsigned long *)&nCompLen, (unsigned char *)pFileBuf, nLength);

			free(pFileBuf);
			pFileBuf = pCompBuf;
			nLength = nCompLen;
		}
	}

	/* Attach the response code and headers to the send buffer */
	Net_AttachRespAndHeaders(pMsgInfo, &daBuffer, nLength);

#ifndef BUILD_STANDARD
	if (pReqInfo->bUseEncrypt == 1)
	{
		GetRandomNumber(IV, CRYPTO_HEADER_IV_SIZE);

		IV[CRYPTO_HEADER_IV_SIZE] = 0;

		DA_Initialize(&daEncBuf);
		DA_Assign(&daEncBuf, (char *)IV, CRYPTO_HEADER_IV_SIZE);

		sprintf(MsgLen, "%d", daBuffer.nCurLen);
		for (i = strlen(MsgLen); i < CRYPTO_HEADER_LENGTH_SIZE; i++)
			MsgLen[i] = PAD_CHAR;

		MsgLen[CRYPTO_HEADER_LENGTH_SIZE] = 0;

		daTemp = Encrypt(MsgLen, CRYPTO_HEADER_LENGTH_SIZE, pReqInfo->bfKey, IV);
		DA_Append(&daEncBuf, daTemp->szArray, daTemp->nCurLen);

		DA_Reset(daTemp);
		free(daTemp);

		daTemp = Encrypt(daBuffer.szArray, daBuffer.nCurLen, pReqInfo->bfKey, IV);
		DA_Append(&daEncBuf, daTemp->szArray, daTemp->nCurLen);

		DA_Reset(daTemp);
		free(daTemp);

		DA_Assign(&daBuffer, daEncBuf.szArray, daEncBuf.nCurLen);
	}
#endif /* !BUILD_STANDARD */

	/* Send the response code and the headers to the server*/
	nRetVal = Net_Write(nSocket, daBuffer.szArray, daBuffer.nCurLen);


#ifndef BUILD_STANDARD
	if ((pReqInfo->bUseEncrypt == 0) && (pMsgInfo->bUseCompression == 0))
#else
	if (pMsgInfo->bUseCompression == 0)
#endif /* !BUILD_STANDARD */
	{
		/* Send the file in chunks */
		numBytes = 1;
		while (nRetVal >= 0 && numBytes > 0)
		{
			numBytes = fread(fileBuf, 1, NET_BUF_SIZE, pFile);
			if (numBytes > 0)
			{
				nRetVal = Net_Write(nSocket, fileBuf, numBytes);
			}
			
		}
	} else {
#ifndef BUILD_STANDARD
		if (pReqInfo->bUseEncrypt != 0)
		{
			GetRandomNumber(IV, CRYPTO_HEADER_IV_SIZE);
			
			IV[CRYPTO_HEADER_IV_SIZE] = 0;
			
			DA_Reset(&daEncBuf);
			DA_Assign(&daEncBuf, (char *)IV, CRYPTO_HEADER_IV_SIZE);
			
			sprintf(MsgLen, "%d", nLength);
			for (i = strlen(MsgLen); i < CRYPTO_HEADER_LENGTH_SIZE; i++)
				MsgLen[i] = PAD_CHAR;
			
			MsgLen[CRYPTO_HEADER_LENGTH_SIZE] = 0;
			
			daTemp = Encrypt(MsgLen, CRYPTO_HEADER_LENGTH_SIZE, pReqInfo->bfKey, IV);
			DA_Append(&daEncBuf, daTemp->szArray, daTemp->nCurLen);
			
			DA_Reset(daTemp);
			free(daTemp);
			
			nSizeToEncrypt = nLength;
			
			if (nSizeToEncrypt < CRYPTO_BLOCK_SIZE)
			{
				nSizeToEncrypt = CRYPTO_BLOCK_SIZE;
			}
			else
			{
				remainder = nSizeToEncrypt % CRYPTO_BLOCK_SIZE;
				
				if (remainder > 0)
					nSizeToEncrypt += CRYPTO_BLOCK_SIZE - remainder;
			}

			pEncBuf = (char *)malloc(nSizeToEncrypt);
			if (pEncBuf == NULL)
			{
				printf("Error encrypting file\n");
				return -1;
			}

			memcpy(pEncBuf, pFileBuf, nLength);

			for (i = nLength; i < nSizeToEncrypt; i++)
				pEncBuf[i] = PAD_CHAR;
			
			daTemp = Encrypt(pEncBuf, nSizeToEncrypt, pReqInfo->bfKey, IV);
			
			DA_Append(&daEncBuf, daTemp->szArray, nSizeToEncrypt);
			
			DA_Reset(daTemp);
			free(daTemp);
			
			nRetVal = Net_Write(nSocket, daEncBuf.szArray, daEncBuf.nCurLen);
		} else {
#endif /* !BUILD_STANDARD */

			nRetVal = Net_Write(nSocket, pFileBuf, nLength);

#ifndef BUILD_STANDARD
		}
#endif /* !BUILD_STANDARD */
	}

	/* Close the file */
	if (fclose(pFile) != 0)
	{
		printf("Could not close file %s.  Error Num=%d\n", szFileName, errno);
		return -1;
	}

	return nRetVal;

}


/*------------------------------------------------------
	Prints program supplied error message, and looks up
	and prints system last error code.

	Returns: Nothing.
-------------------------------------------------------*/
void Net_Error(char *ErrMsg)
{
/* TODO: Figure out way to print the error message 
	associated with the error code */
	printf("%s. Error: %d %s\n", ErrMsg,
#ifdef WIN32
	WSAGetLastError(), "");
#else
	errno, strerror(errno));
#endif
}





int Net_ReadHeaders(int nInSocket, DArray *daInArray, RequestInfo *pReqInfo)
{
	int bContinue = 1;
	int nLen = 0;
	char szBuf[1];		/* Buffer that receives single input char */
	int nRespLen = 0;
	int i = 0;
	int nTotalBytesRead = 0;

#ifndef BUILD_STANDARD

	char IV[CRYPTO_HEADER_IV_SIZE + 1];
	char szLengthEnc[CRYPTO_HEADER_LENGTH_SIZE + 1];
	char szLength[CRYPTO_HEADER_LENGTH_SIZE + 1];
	int nSizeToDecrypt = 0;
	int remainder;
	char *pEncBlock;
	char *pHeaders;


	if (pReqInfo->bUseEncrypt == 0)
	{
#endif /* !BUILD_STANDARD */
		/*  Read 1 byte at a time from the socket, until we hit
		a blank line */
		while (bContinue)
		{
			nLen =  Net_Read (nInSocket, szBuf, 1);
			if (nLen <= 0)
			{
				Net_Error("Net error during attempt to read headers. (RH1)");
				return -1 ;
			} 
			/* append char to the end of the array */
			DA_Append(daInArray, szBuf, 1);
			
			if (EndsWithIgnoreCase(daInArray->szArray, "\r\n\r\n"))
			{
				bContinue = 0;
			} else {
				nRespLen++;
			}
		}
#ifndef BUILD_STANDARD
	} else {
		nLen = Net_Read(nInSocket, IV, CRYPTO_HEADER_IV_SIZE);
		if (nLen < CRYPTO_HEADER_IV_SIZE)
		{
			Net_Error("Net error during attempt to read headers. (RH2)");
			return -1 ;
		} 

		nLen = Net_Read(nInSocket, szLengthEnc, CRYPTO_HEADER_LENGTH_SIZE);
		if (nLen < CRYPTO_HEADER_LENGTH_SIZE)
		{
			Net_Error("Net error during attempt to read headers. (RH3)");
			return -1 ;
		} 

		Decrypt(szLengthEnc, szLength, CRYPTO_HEADER_LENGTH_SIZE, pReqInfo->bfKey, IV); 

		while ((i < CRYPTO_HEADER_LENGTH_SIZE) && (szLength[i] != PAD_CHAR))
			i++;
		
		szLength[i] = '\0';
		
		nRespLen = atoi(szLength);
		
		/* determine how many total bytes will be arriving including the pad? */
		
		nSizeToDecrypt = nRespLen;
		
		if (nSizeToDecrypt < CRYPTO_BLOCK_SIZE)
		{
			nSizeToDecrypt = CRYPTO_BLOCK_SIZE;
		}
		else
		{
			remainder = nSizeToDecrypt % CRYPTO_BLOCK_SIZE;
			
			if (remainder > 0)
				nSizeToDecrypt += CRYPTO_BLOCK_SIZE - remainder;
		}

		pEncBlock = (char *)malloc(nSizeToDecrypt);
		pHeaders = (char *)malloc(nSizeToDecrypt);

		while (nTotalBytesRead < nSizeToDecrypt)
		{
			nLen = Net_Read(nInSocket, pEncBlock+nTotalBytesRead, 
				nSizeToDecrypt-nTotalBytesRead);
			if (nLen < 0)
			{
				Net_Error("Net error during attempt to read message. (RH4)");
				return -1;
			} 
			
			nTotalBytesRead += nLen;
		}

		Decrypt(pEncBlock, pHeaders, nSizeToDecrypt, pReqInfo->bfKey, IV);

		DA_Append(daInArray, pHeaders, nRespLen);
	}
#endif /* !BUILD_STANDARD */

	return nRespLen;

}


/*------------------------------------------------------
	Wrapper for BSD socket read function
-------------------------------------------------------*/
int Net_Read(int socket, char *buf, int len)
{
	return recv(socket, buf, len, 0);
}

/*------------------------------------------------------
	Wrapper for BSD socket write function
-------------------------------------------------------*/
int Net_Write(int socket, char *buf, int len)
{
	return send(socket, buf, len, 0);
}
